# -*- coding: utf-8 -*-
"""
    emmett.orm.migrations.operations
    --------------------------------

    Provides operations handlers for migrations.

    :copyright: 2014 Giovanni Barillari

    Based on the code of Alembic (https://bitbucket.org/zzzeek/alembic)
    :copyright: (c) 2009-2015 by Michael Bayer

    :license: BSD-3-Clause
"""

from __future__ import annotations

import re

from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple

from .base import Migration, Column
from .helpers import DEFAULT_VALUE

if TYPE_CHECKING:
    from .engine import MetaEngine
    from .generation import MetaTable, MetaIndex, MetaForeignKey


class Operation:
    def _env_load_(self, engine: MetaEngine):
        self.engine = engine

    def reverse(self) -> Operation:
        raise NotImplementedError

    def run(self):
        pass


class OpContainer(Operation):
    #: represent a sequence of operations
    def __init__(self, ops: List[Operation] = []):
        self.ops = ops

    def is_empty(self) -> bool:
        return not self.ops

    def as_diffs(self):
        return list(OpContainer._ops_as_diffs(self))

    @classmethod
    def _ops_as_diffs(cls, migrations):
        for op in migrations.ops:
            if hasattr(op, 'ops'):
                for sub_op in cls._ops_as_diffs(op):
                    yield sub_op
            else:
                yield op.to_diff_tuple()


class ModifyTableOps(OpContainer):
    #: a sequence of operations that all apply to a single Table
    def __init__(self, table_name: str, ops: List[Operation]):
        super().__init__(ops)
        self.table_name = table_name

    def reverse(self) -> ModifyTableOps:
        return ModifyTableOps(
            self.table_name,
            ops=list(reversed([op.reverse() for op in self.ops]))
        )


class UpgradeOps(OpContainer):
    #: contains a sequence of operations that would apply during upgrade
    def __init__(self, ops: List[Operation] = [], upgrade_token: str = "upgrades"):
        super().__init__(ops=ops)
        self.upgrade_token = upgrade_token

    def reverse(self) -> DowngradeOps:
        return DowngradeOps(
            ops=list(reversed([op.reverse() for op in self.ops]))
        )


class DowngradeOps(OpContainer):
    #: contains a sequence of operations that would apply during downgrade
    def __init__(self, ops: List[Operation] = [], downgrade_token: str = "downgrades"):
        super().__init__(ops=ops)
        self.downgrade_token = downgrade_token

    def reverse(self):
        return UpgradeOps(
            ops=list(reversed([op.reverse() for op in self.ops]))
        )


class MigrationOp(Operation):
    def __init__(
        self,
        rev_id: str,
        upgrade_ops: UpgradeOps,
        downgrade_ops: DowngradeOps,
        message: Optional[str] = None,
        head: Optional[str] = None,
        splice: Any = None
    ):
        self.rev_id = rev_id
        self.message = message
        self.head = head
        self.splice = splice
        self.upgrade_ops = upgrade_ops
        self.downgrade_ops = downgrade_ops


@Migration.register_operation("create_table")
class CreateTableOp(Operation):
    def __init__(
        self,
        table_name: str,
        columns: List[Column],
        primary_keys: List[str] = [],
        _orig_table: Optional[MetaTable] = None,
        **kw: Any
    ):
        self.table_name = table_name
        self.columns = columns
        self.primary_keys = primary_keys
        self.kw = kw
        self._orig_table = _orig_table

    def reverse(self) -> DropTableOp:
        return DropTableOp.from_table(self.to_table())

    def to_diff_tuple(self) -> Tuple[str, MetaTable]:
        return ("add_table", self.to_table())

    @classmethod
    def from_table(cls, table: MetaTable) -> CreateTableOp:
        return cls(
            table.name,
            [table[colname] for colname in table.fields],
            list(table.primary_keys),
            _orig_table=table
        )

    def to_table(self, migration_context: Any = None) -> MetaTable:
        if self._orig_table is not None:
            return self._orig_table
        from .generation import MetaTable
        return MetaTable(
            self.table_name,
            self.columns,
            self.primary_keys,
            **self.kw
        )

    @classmethod
    def create_table(
        cls,
        table_name: str,
        *columns: Column,
        **kw: Any
    ) -> CreateTableOp:
        return cls(table_name, columns, **kw)

    def run(self):
        self.engine.create_table(
            self.table_name, self.columns, self.primary_keys, **self.kw
        )


@Migration.register_operation("drop_table")
class DropTableOp(Operation):
    def __init__(
        self,
        table_name: str,
        table_kw: Optional[Dict[str, Any]] = None,
        _orig_table: Optional[MetaTable] = None
    ):
        self.table_name = table_name
        self.table_kw = table_kw or {}
        self._orig_table = _orig_table

    def to_diff_tuple(self) -> Tuple[str, MetaTable]:
        return ("remove_table", self.to_table())

    def reverse(self) -> CreateTableOp:
        if self._orig_table is None:
            raise ValueError(
                "operation is not reversible; original table is not present"
            )
        return CreateTableOp.from_table(self._orig_table)

    @classmethod
    def from_table(cls, table: MetaTable) -> DropTableOp:
        return cls(table.name, _orig_table=table)

    def to_table(self) -> MetaTable:
        if self._orig_table is not None:
            return self._orig_table
        from .generation import MetaTable
        return MetaTable(
            self.table_name,
            **self.table_kw
        )

    @classmethod
    def drop_table(cls, table_name: str, **kw: Any) -> DropTableOp:
        return cls(table_name, table_kw=kw)

    def run(self):
        self.engine.drop_table(self.table_name)


class AlterTableOp(Operation):
    def __init__(self, table_name: str):
        self.table_name = table_name


@Migration.register_operation("rename_table")
class RenameTableOp(AlterTableOp):
    def __init__(self, old_table_name: str, new_table_name: str):
        super().__init__(old_table_name)
        self.new_table_name = new_table_name

    @classmethod
    def rename_table(cls, old_table_name: str, new_table_name: str) -> RenameTableOp:
        return cls(old_table_name, new_table_name)

    def run(self):
        raise NotImplementedError('Table renaming is currently not supported.')


@Migration.register_operation("add_column")
class AddColumnOp(AlterTableOp):
    def __init__(self, table_name: str, column: Column):
        super().__init__(table_name)
        self.column = column

    def reverse(self) -> DropColumnOp:
        return DropColumnOp.from_column_and_tablename(self.table_name, self.column)

    def to_diff_tuple(self) -> Tuple[str, str, Column]:
        return ("add_column", self.table_name, self.column)

    def to_column(self) -> Column:
        return self.column

    @classmethod
    def from_column_and_tablename(cls, tname: str, col: Column) -> AddColumnOp:
        return cls(tname, col)

    @classmethod
    def add_column(cls, table_name: str, column: Column) -> AddColumnOp:
        return cls(table_name, column)

    def run(self):
        self.engine.add_column(self.table_name, self.column)


@Migration.register_operation("drop_column")
class DropColumnOp(AlterTableOp):
    def __init__(
        self,
        table_name: str,
        column_name: str,
        _orig_column: Optional[Column] = None,
        **kw: Any
    ):
        super().__init__(table_name)
        self.column_name = column_name
        self.kw = kw
        self._orig_column = _orig_column

    def to_diff_tuple(self) -> Tuple[str, str, Column]:
        return ("remove_column", self.table_name, self.to_column())

    def reverse(self) -> AddColumnOp:
        if self._orig_column is None:
            raise ValueError(
                "operation is not reversible; original column is not present"
            )

        return AddColumnOp.from_column_and_tablename(
            self.table_name, self._orig_column
        )

    @classmethod
    def from_column_and_tablename(cls, tname: str, col: Column) -> DropColumnOp:
        return cls(tname, col.name, _orig_column=col)

    def to_column(self) -> Column:
        if self._orig_column is not None:
            return self._orig_column
        return Column(self.column_name, **self.kw)

    @classmethod
    def drop_column(cls, table_name: str, column_name: str, **kw: Any) -> DropColumnOp:
        return cls(table_name, column_name, **kw)

    def run(self):
        self.engine.drop_column(self.table_name, self.column_name)


@Migration.register_operation("alter_column")
class AlterColumnOp(AlterTableOp):
    def __init__(
        self,
        table_name: str,
        column_name: str,
        existing_type: Optional[str] = None,
        existing_length: Optional[int] = None,
        existing_default: Any = None,
        existing_notnull: Optional[bool] = None,
        modify_notnull: Optional[bool] = None,
        modify_default: Any = DEFAULT_VALUE,
        modify_name: Optional[str] = None,
        modify_type: Optional[str] = None,
        modify_length: Optional[int] = None,
        **kw: Any
    ):
        super().__init__(table_name)
        self.column_name = column_name
        self.existing_type = existing_type
        self.existing_length = existing_length
        self.existing_default = existing_default
        self.existing_notnull = existing_notnull
        self.modify_notnull = modify_notnull
        self.modify_default = modify_default
        self.modify_name = modify_name
        self.modify_type = modify_type
        self.modify_length = modify_length
        self.kw = kw

    def to_diff_tuple(self) -> List[Tuple[str, str, str, Dict[str, Any], Any, Any]]:
        col_diff = []
        tname, cname = self.table_name, self.column_name

        if self.modify_type is not None:
            col_diff.append(
                (
                    "modify_type",
                    tname,
                    cname,
                    {
                        "existing_length": self.existing_length,
                        "existing_notnull": self.existing_notnull,
                        "existing_default": self.existing_default,
                        **{
                            nkey: nval for nkey, nval in self.kw.items()
                            if nkey.startswith('existing_')
                        }
                    },
                    self.existing_type,
                    self.modify_type
                )
            )

        if self.modify_length is not None:
            col_diff.append(
                (
                    "modify_length",
                    tname,
                    cname,
                    {
                        "existing_type": self.existing_type,
                        "existing_notnull": self.existing_notnull,
                        "existing_default": self.existing_default
                    },
                    self.existing_length,
                    self.modify_length
                )
            )

        if self.modify_notnull is not None:
            col_diff.append(
                (
                    "modify_notnull",
                    tname,
                    cname,
                    {
                        "existing_type": self.existing_type,
                        "existing_default": self.existing_default
                    },
                    self.existing_notnull,
                    self.modify_notnull
                )
            )

        if self.modify_default is not DEFAULT_VALUE:
            col_diff.append(
                (
                    "modify_default",
                    tname,
                    cname,
                    {
                        "existing_notnull": self.existing_notnull,
                        "existing_type": self.existing_type
                    },
                    self.existing_default,
                    self.modify_default
                )
            )

        for key, val in self.kw.items():
            if key.startswith("modify_"):
                attr = key.split("modify_")[-1]
                col_diff.append(
                    (
                        key,
                        tname,
                        cname,
                        {
                            "existing_type": self.existing_type,
                            **{
                                nkey: nval for nkey, nval in self.kw.items()
                                if nkey.startswith('existing_')
                            }
                        },
                        self.kw.get(f"existing_{attr}"),
                        val
                    )
                )

        return col_diff

    def has_changes(self) -> bool:
        hc = (
            self.modify_notnull is not None or
            self.modify_default is not DEFAULT_VALUE or
            self.modify_type is not None or
            self.modify_length is not None
        )
        if hc:
            return True
        for kw in self.kw:
            if kw.startswith('modify_'):
                return True
        return False

    def reverse(self) -> AlterColumnOp:
        kw = self.kw.copy()
        kw['existing_type'] = self.existing_type
        kw['existing_length'] = self.existing_length
        kw['existing_notnull'] = self.existing_notnull
        kw['existing_default'] = self.existing_default
        if self.modify_type is not None:
            kw['modify_type'] = self.modify_type
        if self.modify_length is not None:
            kw['modify_length'] = self.modify_length
        if self.modify_notnull is not None:
            kw['modify_notnull'] = self.modify_notnull
        if self.modify_default is not DEFAULT_VALUE:
            kw['modify_default'] = self.modify_default

        all_keys = set(m.group(1) for m in [
            re.match(r'^(?:existing_|modify_)(.+)$', k)
            for k in kw
        ] if m)

        for k in all_keys:
            if 'modify_%s' % k in kw:
                swap = kw['existing_%s' % k]
                kw['existing_%s' % k] = kw['modify_%s' % k]
                kw['modify_%s' % k] = swap

        return self.__class__(
            self.table_name, self.column_name, **kw
        )

    @classmethod
    def alter_column(
        cls,
        table_name: str,
        column_name: str,
        notnull: Optional[bool] = None,
        default: Any = DEFAULT_VALUE,
        new_column_name: Optional[str] = None,
        type: Optional[str] = None,
        length: Optional[int] = None,
        existing_type: Optional[str] = None,
        existing_length: Optional[int] = None,
        existing_default: Any = None,
        existing_notnull: Optional[bool] = None,
        **kw: Any
    ) -> AlterColumnOp:
        return cls(
            table_name,
            column_name,
            existing_type=existing_type,
            existing_length=existing_length,
            existing_default=existing_default,
            existing_notnull=existing_notnull,
            modify_name=new_column_name,
            modify_type=type,
            modify_length=length,
            modify_default=default,
            modify_notnull=notnull,
            **kw
        )

    def run(self):
        self.engine.alter_column(
            self.table_name, self.column_name, self.to_diff_tuple()
        )


@Migration.register_operation("create_index")
class CreateIndexOp(Operation):
    def __init__(
        self,
        index_name: str,
        table_name: str,
        fields: List[str] = [],
        expressions: List[str] = [],
        unique: bool = False,
        _orig_index: Optional[MetaIndex] = None,
        **kw: Any
    ):
        self.index_name = index_name
        self.table_name = table_name
        self.fields = fields
        self.expressions = expressions
        self.unique = unique
        self.kw = kw
        self._orig_index = _orig_index

    def reverse(self) -> DropIndexOp:
        return DropIndexOp.from_index(self.to_index())

    def to_diff_tuple(self) -> Tuple[str, MetaIndex]:
        return ("create_index", self.to_index())

    @classmethod
    def from_index(cls, index: MetaIndex) -> CreateIndexOp:
        return cls(
            index.name, index.table_name, index.fields, index.expressions,
            index.unique, _orig_index=index, **index.kw
        )

    def to_index(self) -> MetaIndex:
        if self._orig_index is not None:
            return self._orig_index
        from .generation import MetaIndex
        return MetaIndex(
            self.table_name,
            self.index_name,
            self.fields,
            self.expressions,
            self.unique,
            **self.kw
        )

    @classmethod
    def create_index(
        cls,
        index_name: str,
        table_name: str,
        fields: List[str] = [],
        expressions: List[str] = [],
        unique: bool = False,
        **kw: Any
    ) -> CreateIndexOp:
        return cls(index_name, table_name, fields, expressions, unique, **kw)

    def run(self):
        self.engine.create_index(
            self.index_name,
            self.table_name,
            self.fields,
            self.expressions,
            self.unique,
            **self.kw
        )


@Migration.register_operation("drop_index")
class DropIndexOp(Operation):
    def __init__(
        self,
        index_name: str,
        table_name: Optional[str] = None,
        _orig_index: Optional[MetaIndex] = None
    ):
        self.index_name = index_name
        self.table_name = table_name
        self._orig_index = _orig_index

    def to_diff_tuple(self) -> Tuple[str, MetaIndex]:
        return ("remove_index", self.to_index())

    def reverse(self) -> CreateIndexOp:
        if self._orig_index is None:
            raise ValueError(
                "operation is not reversible; original index is not present"
            )
        return CreateIndexOp.from_index(self._orig_index)

    @classmethod
    def from_index(cls, index: MetaIndex) -> DropIndexOp:
        return cls(index.name, index.table_name, index)

    def to_index(self) -> MetaIndex:
        if self._orig_index is not None:
            return self._orig_index
        from .generation import MetaIndex
        return MetaIndex(self.table_name, self.index_name, [], [], False)

    @classmethod
    def drop_index(cls, index_name: str, table_name: str) -> DropIndexOp:
        return cls(index_name, table_name)

    def run(self):
        self.engine.drop_index(self.index_name, self.table_name)


@Migration.register_operation("create_foreign_key")
class CreateForeignKeyConstraintOp(AlterTableOp):
    def __init__(
        self,
        name: str,
        table_name: str,
        foreign_table_name: str,
        column_names: List[str],
        foreign_keys: List[str],
        on_delete: str,
        _orig_fk: Optional[MetaForeignKey] = None,
        **kw: Any
    ):
        super().__init__(table_name)
        self.constraint_name = name
        self.foreign_table_name = foreign_table_name
        self.column_names = column_names
        self.foreign_keys = foreign_keys
        self.on_delete = on_delete
        self.kw = kw
        self._orig_fk = _orig_fk
        if len(self.column_names) != len(self.foreign_keys):
            raise SyntaxError("local and foreign columns number should match")

    def reverse(self) -> DropForeignKeyConstraintOp:
        return DropForeignKeyConstraintOp.from_foreign_key(self.to_foreign_key())

    def to_diff_tuple(self) -> Tuple[str, MetaForeignKey]:
        return ("create_fk_constraint", self.to_foreign_key())

    @classmethod
    def from_foreign_key(
        cls,
        foreign_key: MetaForeignKey
    ) -> CreateForeignKeyConstraintOp:
        return cls(
            foreign_key.name,
            foreign_key.table_name,
            foreign_key.foreign_table_name,
            foreign_key.column_names,
            foreign_key.foreign_keys,
            foreign_key.on_delete,
            _orig_fk=foreign_key
        )

    def to_foreign_key(self) -> MetaForeignKey:
        if self._orig_fk is not None:
            return self._orig_fk

        from .generation import MetaForeignKey
        return MetaForeignKey(
            self.table_name,
            self.constraint_name,
            self.column_names,
            self.foreign_table_name,
            self.foreign_keys,
            self.on_delete
        )

    @classmethod
    def create_foreign_key(
        cls,
        name: str,
        table_name: str,
        foreign_table_name: str,
        column_names: List[str],
        foreign_keys: List[str],
        on_delete: str
    ) -> CreateForeignKeyConstraintOp:
        return cls(
            name=name,
            table_name=table_name,
            foreign_table_name=foreign_table_name,
            column_names=column_names,
            foreign_keys=foreign_keys,
            on_delete=on_delete
        )

    def run(self):
        self.engine.create_foreign_key_constraint(
            self.constraint_name,
            self.table_name,
            self.column_names,
            self.foreign_table_name,
            self.foreign_keys,
            self.on_delete
        )


@Migration.register_operation("drop_foreign_key")
class DropForeignKeyConstraintOp(AlterTableOp):
    def __init__(
        self,
        name: str,
        table_name: str,
        _orig_fk: Optional[MetaForeignKey] = None,
        **kw: Any
    ):
        super().__init__(table_name)
        self.constraint_name = name
        self.kw = kw
        self._orig_fk = _orig_fk

    def reverse(self) -> CreateForeignKeyConstraintOp:
        if self._orig_fk is None:
            raise ValueError(
                "operation is not reversible; original constraint is not present"
            )
        return CreateForeignKeyConstraintOp.from_foreign_key(self._orig_fk)

    def to_diff_tuple(self) -> Tuple[str, MetaForeignKey]:
        return ("drop_fk_constraint", self.to_foreign_key())

    @classmethod
    def from_foreign_key(
        cls,
        foreign_key: MetaForeignKey
    ) -> DropForeignKeyConstraintOp:
        return cls(
            foreign_key.name,
            foreign_key.table_name,
            _orig_fk=foreign_key
        )

    def to_foreign_key(self):
        if self._orig_fk is not None:
            return self._orig_fk

        from .generation import MetaForeignKey
        return MetaForeignKey(self.table_name, self.constraint_name, [], '', [], '')

    @classmethod
    def drop_foreign_key(
        cls,
        name: str,
        table_name: str
    ) -> DropForeignKeyConstraintOp:
        return DropForeignKeyConstraintOp(name, table_name)

    def run(self):
        self.engine.drop_foreign_key_constraint(self.constraint_name, self.table_name)


# @Migration.register_operation("execute")
# class ExecuteSQLOp(Operation):
#     def __init__(self, sqltext):
#         self.sqltext = sqltext

#     def run(self):
#         pass



# -*- coding: utf-8 -*-
"""
    emmett.rsgi.handlers
    --------------------

    Provides RSGI handlers.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from __future__ import annotations

import asyncio
import os
import re

from typing import Awaitable, Callable, Optional, Tuple

from granian.rsgi import (
    HTTPProtocol,
    ProtocolClosed,
    Scope,
    WebsocketMessageType,
    WebsocketProtocol
)

from ..ctx import RequestContext, WSContext, current
from ..debug import smart_traceback, debug_handler
from ..http import HTTPResponse, HTTPFile, HTTP
from ..utils import cachedprop
from ..wrappers.response import Response

from .helpers import WSTransport
from .wrappers import Request, Websocket

REGEX_STATIC = re.compile(
    r'^/static/(?P<m>__[\w\-\.]+__/)?(?P<v>_\d+\.\d+\.\d+/)?(?P<f>.*?)$'
)
REGEX_STATIC_LANG = re.compile(
    r'^/(?P<l>\w{2}/)?static/(?P<m>__[\w\-\.]__+/)?(?P<v>_\d+\.\d+\.\d+/)?(?P<f>.*?)$'
)


class Handler:
    __slots__ = ['app']

    def __init__(self, app):
        self.app = app


class RequestHandler(Handler):
    __slots__ = ['router']

    def __init__(self, app):
        super().__init__(app)
        self._bind_router()
        self._configure_methods()

    def _bind_router(self):
        raise NotImplementedError

    def _configure_methods(self):
        raise NotImplementedError


class HTTPHandler(RequestHandler):
    __slots__ = ['pre_handler', 'static_handler', 'static_matcher', '__dict__']

    def _bind_router(self):
        self.router = self.app._router_http

    def _configure_methods(self):
        self.static_matcher = (
            self._static_lang_matcher if self.app.language_force_on_url else
            self._static_nolang_matcher
        )
        self.static_handler = (
            self._static_handler if self.app.config.handle_static else
            self.dynamic_handler
        )
        self.pre_handler = (
            self._prefix_handler if self.router._prefix_main else
            self.static_handler
        )

    async def __call__(
        self,
        scope: Scope,
        protocol: HTTPProtocol
    ):
        try:
            http = await self.pre_handler(scope, protocol, scope.path)
        except asyncio.TimeoutError:
            self.app.log.warn(
                f"Timeout sending response: ({scope.path})"
            )
        if coro := http.rsgi(protocol):
            await coro

    @cachedprop
    def error_handler(self) -> Callable[[], Awaitable[str]]:
        return (
            self._debug_handler if self.app.debug else self.exception_handler
        )

    @cachedprop
    def exception_handler(self) -> Callable[[], Awaitable[str]]:
        return self.app.error_handlers.get(500, self._exception_handler)

    @staticmethod
    async def _http_response(code: int) -> HTTPResponse:
        return HTTP(code)

    def _prefix_handler(
        self,
        scope: Scope,
        protocol: HTTPProtocol,
        path: str
    ) -> Awaitable[HTTPResponse]:
        if not path.startswith(self.router._prefix_main):
            return self._http_response(404)
        path = path[self.router._prefix_main_len:] or '/'
        return self.static_handler(scope, protocol, path)

    def _static_lang_matcher(
        self, path: str
    ) -> Tuple[Optional[str], Optional[str]]:
        match = REGEX_STATIC_LANG.match(path)
        if match:
            lang, mname, version, file_name = match.group('l', 'm', 'v', 'f')
            if mname:
                mod = self.app._modules.get(mname)
                spath = mod._static_path if mod else self.app.static_path
            else:
                spath = self.app.static_path
            static_file = os.path.join(spath, file_name)
            if lang:
                lang_file = os.path.join(spath, lang, file_name)
                if os.path.exists(lang_file):
                    static_file = lang_file
            return static_file, version
        return None, None

    def _static_nolang_matcher(
        self, path: str
    ) -> Tuple[Optional[str], Optional[str]]:
        if path.startswith('/static/'):
            mname, version, file_name = REGEX_STATIC.match(path).group('m', 'v', 'f')
            if mname:
                mod = self.app._modules.get(mname[2:-3])
                static_file = os.path.join(mod._static_path, file_name) if mod else None
            elif file_name:
                static_file = os.path.join(self.app.static_path, file_name)
            else:
                static_file = None
            return static_file, version
        return None, None

    async def _static_response(self, file_path: str) -> HTTPFile:
        return HTTPFile(file_path)

    def _static_handler(
        self,
        scope: Scope,
        protocol: HTTPProtocol,
        path: str
    ) -> Awaitable[HTTPResponse]:
        #: handle internal assets
        if path.startswith('/__emmett__'):
            file_name = path[12:]
            if not file_name:
                return self._http_response(404)
            static_file = os.path.join(
                os.path.dirname(__file__), '..', 'assets', file_name
            )
            if os.path.splitext(static_file)[1] == 'html':
                return self._http_response(404)
            return self._static_response(static_file)
        #: handle app assets
        static_file, _ = self.static_matcher(path)
        if static_file:
            return self._static_response(static_file)
        return self.dynamic_handler(scope, protocol, path)

    async def dynamic_handler(
        self,
        scope: Scope,
        protocol: HTTPProtocol,
        path: str
    ) -> HTTPResponse:
        request = Request(
            scope,
            path,
            protocol,
            max_content_length=self.app.config.request_max_content_length,
            body_timeout=self.app.config.request_body_timeout
        )
        response = Response()
        ctx = RequestContext(self.app, request, response)
        ctx_token = current._init_(ctx)
        try:
            http = await self.router.dispatch(request, response)
        except HTTPResponse as http_exception:
            http = http_exception
            #: render error with handlers if in app
            error_handler = self.app.error_handlers.get(http.status_code)
            if error_handler:
                http = HTTP(
                    http.status_code,
                    await error_handler(),
                    headers=response.headers,
                    cookies=response.cookies
                )
        except Exception:
            self.app.log.exception('Application exception:')
            http = HTTP(
                500,
                await self.error_handler(),
                headers=response.headers
            )
        finally:
            current._close_(ctx_token)
        return http

    async def _debug_handler(self) -> str:
        current.response.headers._data['content-type'] = (
            'text/html; charset=utf-8'
        )
        return debug_handler(smart_traceback(self.app))

    async def _exception_handler(self) -> str:
        current.response.headers._data['content-type'] = 'text/plain'
        return 'Internal error'


class WSHandler(RequestHandler):
    __slots__ = ['pre_handler', '__dict__']

    def _bind_router(self):
        self.router = self.app._router_ws

    def _configure_methods(self):
        self.pre_handler = (
            self._prefix_handler if self.router._prefix_main else
            self.dynamic_handler
        )

    async def __call__(
        self,
        scope: Scope,
        protocol: WebsocketProtocol
    ):
        transport = WSTransport(protocol)
        task_transport = asyncio.create_task(self.handle_transport(transport))
        task_request = asyncio.create_task(self.handle_request(scope, transport))
        _, pending = await asyncio.wait(
            [task_request, task_transport], return_when=asyncio.FIRST_COMPLETED
        )
        for task in pending:
            task.cancel()
        self._close_connection(transport)

    async def handle_transport(self, transport: WSTransport):
        await transport.accepted.wait()
        try:
            while True:
                msg = await transport.transport.receive()
                if msg.kind == WebsocketMessageType.close:
                    transport.interrupted = True
                    break
                await transport.input.put(msg)
        except ProtocolClosed:
            transport.interrupted = True

    def handle_request(
        self,
        scope: Scope,
        transport: WSTransport
    ):
        return self.pre_handler(scope, transport, scope.path)

    async def _empty_awaitable(self):
        return

    def _prefix_handler(
        self,
        scope: Scope,
        transport: WSTransport,
        path: str
    ) -> Awaitable[None]:
        if not path.startswith(self.router._prefix_main):
            transport.status = 404
            return self._empty_awaitable()
        path = path[self.router._prefix_main_len:] or '/'
        return self.dynamic_handler(scope, transport, path)

    async def dynamic_handler(
        self,
        scope: Scope,
        transport: WSTransport,
        path: str
    ):
        ctx = WSContext(self.app, Websocket(scope, path, transport))
        ctx_token = current._init_(ctx)
        try:
            await self.router.dispatch(ctx.websocket)
        except HTTPResponse as http:
            transport.status = http.status_code
        except asyncio.CancelledError:
            if not transport.interrupted:
                self.app.log.exception('Application exception:')
        except Exception:
            transport.status = 500
            self.app.log.exception('Application exception:')
        finally:
            current._close_(ctx_token)

    def _close_connection(self, transport: WSTransport):
        transport.protocol.close(transport.status)



# -*- coding: utf-8 -*-
"""
    emmett.asgi.wrappers
    --------------------

    Provides ASGI request and websocket wrappers

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import asyncio

from datetime import datetime
from typing import Any, Callable, Dict, Iterator, List, Mapping, Optional, Tuple, Union
from urllib.parse import parse_qs

from ..datastructures import sdict
from ..http import HTTP
from ..utils import cachedprop
from ..wrappers.helpers import regex_client
from ..wrappers.request import Request as _Request
from ..wrappers.websocket import Websocket as _Websocket
from .helpers import RequestCancelled
from .typing import Scope, Receive, Send

_push_headers = {
    "accept",
    "accept-encoding",
    "accept-language",
    "cache-control",
    "user-agent"
}


class Headers(Mapping[str, str]):
    __slots__ = ["_data"]

    def __init__(self, scope: Dict[str, Any]):
        self._data: Dict[bytes, bytes] = {
            key: val for key, val in scope["headers"]
        }

    __hash__ = None  # type: ignore

    def __getitem__(self, key: str) -> str:
        return self._data[key.lower().encode("latin-1")].decode("latin-1")

    def __contains__(self, key: str) -> bool:  # type: ignore
        return key.lower().encode("latin-1") in self._data

    def __iter__(self) -> Iterator[str]:
        for key in self._data.keys():
            yield key.decode("latin-1")

    def __len__(self) -> int:
        return len(self._data)

    def get(
        self,
        key: str,
        default: Optional[Any] = None,
        cast: Optional[Callable[[Any], Any]] = None
    ) -> Any:
        rv = self._data.get(key.lower().encode("latin-1"))
        rv = rv.decode() if rv is not None else default  # type: ignore
        if cast is None:
            return rv
        try:
            return cast(rv)
        except ValueError:
            return default

    def items(self) -> Iterator[Tuple[str, str]]:  # type: ignore
        for key, value in self._data.items():
            yield key.decode("latin-1"), value.decode("latin-1")

    def keys(self) -> Iterator[str]:  # type: ignore
        for key in self._data.keys():
            yield key.decode("latin-1")

    def values(self) -> Iterator[str]:  # type: ignore
        for value in self._data.values():
            yield value.decode("latin-1")


class Body:
    __slots__ = ('_data', '_receive', '_max_content_length')

    def __init__(self, receive, max_content_length=None):
        self._data = bytearray()
        self._receive = receive
        self._max_content_length = max_content_length

    def append(self, data: bytes):
        if data == b'':
            return
        self._data.extend(data)
        if (
            self._max_content_length is not None and
            len(self._data) > self._max_content_length
        ):
            raise HTTP(413, 'Request entity too large')

    async def __load(self) -> bytes:
        while True:
            event = await self._receive()
            if event['type'] == 'http.request':
                self.append(event['body'])
                if not event.get('more_body', False):
                    break
            elif event['type'] == 'http.disconnect':
                raise RequestCancelled
        return bytes(self._data)

    def __await__(self):
        return self.__load().__await__()


class ASGIIngressMixin:
    def __init__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send
    ):
        self._scope = scope
        self._receive = receive
        self._send = send
        self.scheme = scope['scheme']
        self.path = scope['emt.path']

    @cachedprop
    def headers(self) -> Headers:
        return Headers(self._scope)

    @cachedprop
    def query_params(self) -> sdict[str, Union[str, List[str]]]:
        rv: sdict[str, Any] = sdict()
        for key, values in parse_qs(
            self._scope['query_string'].decode('latin-1'), keep_blank_values=True
        ).items():
            if len(values) == 1:
                rv[key] = values[0]
                continue
            rv[key] = values
        return rv

    @cachedprop
    def client(self) -> str:
        g = regex_client.search(self.headers.get('x-forwarded-for', ''))
        client = (
            (g.group() or '').split(',')[0] if g else (
                self._scope['client'][0] if self._scope['client'] else None
            )
        )
        if client in (None, '', 'unknown', 'localhost'):
            client = '::1' if self.host.startswith('[') else '127.0.0.1'
        return client  # type: ignore


class Request(ASGIIngressMixin, _Request):
    __slots__ = ['_scope', '_receive', '_send']

    def __init__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
        max_content_length: Optional[int] = None,
        body_timeout: Optional[int] = None
    ):
        super().__init__(scope, receive, send)
        self.max_content_length = max_content_length
        self.body_timeout = body_timeout
        self._now = datetime.utcnow()
        self.method = scope['method']

    @cachedprop
    def _input(self):
        return Body(self._receive, self.max_content_length)

    @cachedprop
    async def body(self) -> bytes:
        if (
            self.max_content_length and
            self.content_length > self.max_content_length
        ):
            raise HTTP(413, 'Request entity too large')
        try:
            rv = await asyncio.wait_for(self._input, timeout=self.body_timeout)
        except asyncio.TimeoutError:
            raise HTTP(408, 'Request timeout')
        return rv

    async def push_promise(self, path: str):
        if "http.response.push" not in self._scope.get("extensions", {}):
            return
        await self._send({
            "type": "http.response.push",
            "path": path,
            "headers": [
                (key.encode("latin-1"), self.headers[key].encode("latin-1"))
                for key in _push_headers & set(self.headers.keys())
            ]
        })


class Websocket(ASGIIngressMixin, _Websocket):
    __slots__ = ['_scope', '_receive', '_send', '_accepted']

    def __init__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send
    ):
        super().__init__(scope, receive, send)
        self._accepted = False
        self._flow_receive = None
        self._flow_send = None
        self.receive = self._accept_and_receive
        self.send = self._accept_and_send

    @property
    def _asgi_spec_version(self) -> int:
        return int(''.join(
            self._scope.get('asgi', {}).get('spec_version', '2.0').split('.')
        ))

    def _encode_headers(
        self,
        headers: Dict[str, str]
    ) -> List[Tuple[bytes, bytes]]:
        return [
            (key.encode('utf-8'), val.encode('utf-8'))
            for key, val in headers.items()
        ]

    async def accept(
        self,
        headers: Optional[Dict[str, str]] = None,
        subprotocol: Optional[str] = None
    ):
        if self._accepted:
            return
        message: Dict[str, Any] = {
            'type': 'websocket.accept',
            'subprotocol': subprotocol
        }
        if headers and self._asgi_spec_version > 20:
            message['headers'] = self._encode_headers(headers)
        await self._send(message)
        self._accepted = True
        self.receive = self._wrapped_receive
        self.send = self._wrapped_send

    async def _wrapped_receive(self) -> Any:
        data = await self._receive()
        for method in self._flow_receive:  # type: ignore
            data = method(data)
        return data

    async def _wrapped_send(self, data: Any):
        for method in self._flow_send:  # type: ignore
            data = method(data)
        if isinstance(data, str):
            await self._send({'type': 'websocket.send', 'text': data})
        else:
            await self._send({'type': 'websocket.send', 'bytes': data})



# -*- coding: utf-8 -*-
"""
    emmett.orm.models
    -----------------

    Provides model layer for Emmet's ORM.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import operator
import types

from collections import OrderedDict
from functools import reduce
from typing import Any, Callable

from ..datastructures import sdict
from .apis import (
    compute,
    rowattr,
    rowmethod,
    scope,
    belongs_to,
    refers_to,
    has_one,
    has_many
)
from .errors import (
    InsertFailureOnSave,
    SaveException,
    UpdateFailureOnSave,
    ValidationError,
    DestroyException
)
from .helpers import (
    Callback,
    ReferenceData,
    RowReferenceMixin,
    RowReferenceMulti,
    camelize,
    decamelize,
    make_tablename,
    typed_row_reference,
    typed_row_reference_from_record,
    wrap_scope_on_model,
    wrap_virtual_on_model
)
from .objects import Field, StructuredRow
from .wrappers import HasOneWrap, HasOneViaWrap, HasManyWrap, HasManyViaWrap


class MetaModel(type):
    _inheritable_dict_attrs_ = [
        'indexes', 'validation', ('fields_rw', {'id': False}), 'foreign_keys',
        'default_values', 'update_values', 'repr_values',
        'form_labels', 'form_info', 'form_widgets'
    ]

    def __new__(cls, name, bases, attrs):
        new_class = type.__new__(cls, name, bases, attrs)
        #: collect declared attributes
        tablename = attrs.get('tablename')
        fields = []
        vfields = []
        computations = []
        callbacks = []
        declared_fields = OrderedDict()
        declared_vfields = OrderedDict()
        declared_computations = OrderedDict()
        declared_callbacks = OrderedDict()
        declared_scopes = {}
        for key, value in list(attrs.items()):
            if isinstance(value, Field):
                fields.append((key, value))
            elif isinstance(value, rowattr):
                vfields.append((key, value))
            elif isinstance(value, compute):
                computations.append((key, value))
            elif isinstance(value, Callback):
                callbacks.append((key, value))
            elif isinstance(value, scope):
                declared_scopes[key] = value
        declared_relations = sdict(
            belongs=OrderedDict(), refers=OrderedDict(),
            hasone=OrderedDict(), hasmany=OrderedDict()
        )
        for ref in belongs_to._references_.values():
            for item in ref.reference:
                rkey = list(item)[0] if isinstance(item, dict) else item
                declared_relations.belongs[rkey] = item
        belongs_to._references_ = {}
        for ref in refers_to._references_.values():
            for item in ref.reference:
                rkey = list(item)[0] if isinstance(item, dict) else item
                declared_relations.refers[rkey] = item
        refers_to._references_ = {}
        for ref in has_one._references_.values():
            for item in ref.reference:
                rkey = list(item)[0] if isinstance(item, dict) else item
                declared_relations.hasone[rkey] = item
        has_one._references_ = {}
        for ref in has_many._references_.values():
            for item in ref.reference:
                rkey = list(item)[0] if isinstance(item, dict) else item
                declared_relations.hasmany[rkey] = item
        has_many._references_ = {}
        #: sort declared attributes that keeps order
        fields.sort(key=lambda x: x[1]._inst_count_)
        vfields.sort(key=lambda x: x[1]._inst_count_)
        computations.sort(key=lambda x: x[1]._inst_count_)
        callbacks.sort(key=lambda x: x[1]._inst_count_)
        declared_fields.update(fields)
        declared_vfields.update(vfields)
        declared_computations.update(computations)
        declared_callbacks.update(callbacks)
        #: store declared attributes in class
        new_class._declared_tablename_ = tablename
        new_class._declared_fields_ = declared_fields
        new_class._declared_virtuals_ = declared_vfields
        new_class._declared_computations_ = declared_computations
        new_class._declared_callbacks_ = declared_callbacks
        new_class._declared_scopes_ = declared_scopes
        new_class._declared_belongs_ref_ = declared_relations.belongs
        new_class._declared_refers_ref_ = declared_relations.refers
        new_class._declared_hasone_ref_ = declared_relations.hasone
        new_class._declared_hasmany_ref_ = declared_relations.hasmany
        #: get super declared attributes
        all_fields = OrderedDict()
        all_vfields = OrderedDict()
        all_computations = OrderedDict()
        all_callbacks = OrderedDict()
        all_scopes = {}
        all_relations = sdict(
            belongs=OrderedDict(), refers=OrderedDict(),
            hasone=OrderedDict(), hasmany=OrderedDict()
        )
        super_vfields = OrderedDict()
        for base in reversed(new_class.__mro__[1:]):
            if hasattr(base, '_declared_fields_'):
                all_fields.update(base._declared_fields_)
            if hasattr(base, '_declared_virtuals_'):
                all_vfields.update(base._declared_virtuals_)
                super_vfields.update(base._declared_virtuals_)
            if hasattr(base, '_declared_computations_'):
                all_computations.update(base._declared_computations_)
            if hasattr(base, '_declared_callbacks_'):
                all_callbacks.update(base._declared_callbacks_)
            if hasattr(base, '_declared_scopes_'):
                all_scopes.update(base._declared_scopes_)
            for key in list(all_relations):
                attrkey = '_declared_' + key + '_ref_'
                if hasattr(base, attrkey):
                    all_relations[key].update(getattr(base, attrkey))
        #: compose 'all' attributes
        all_fields.update(declared_fields)
        all_vfields.update(declared_vfields)
        all_computations.update(declared_computations)
        all_callbacks.update(declared_callbacks)
        all_scopes.update(declared_scopes)
        for key in list(all_relations):
            all_relations[key].update(declared_relations[key])
        #: store 'all' attributes on class
        new_class._all_fields_ = all_fields
        new_class._all_virtuals_ = all_vfields
        new_class._all_computations_ = all_computations
        new_class._all_callbacks_ = all_callbacks
        new_class._all_scopes_ = all_scopes
        new_class._all_belongs_ref_ = all_relations.belongs
        new_class._all_refers_ref_ = all_relations.refers
        new_class._all_hasone_ref_ = all_relations.hasone
        new_class._all_hasmany_ref_ = all_relations.hasmany
        #: store 'super' attributes on class
        new_class._super_virtuals_ = super_vfields
        return new_class


class Model(metaclass=MetaModel):
    db = None
    table = None

    auto_validation = True

    @classmethod
    def _init_inheritable_dicts_(cls):
        if cls.__bases__ != (object,):
            return
        for attr in cls._inheritable_dict_attrs_:
            if isinstance(attr, tuple):
                attr_name, default = attr
            else:
                attr_name, default = attr, {}
            if not isinstance(default, dict):
                raise SyntaxError(f"{attr_name} is not a dictionary")
            setattr(cls, attr_name, default)

    @classmethod
    def _merge_inheritable_dicts_(cls, models):
        for attr in cls._inheritable_dict_attrs_:
            if isinstance(attr, tuple):
                attr_name = attr[0]
            else:
                attr_name = attr
            attrs = {}
            for model in models:
                superattrs = getattr(model, attr_name)
                for k, v in superattrs.items():
                    attrs[k] = v
            for k, v in getattr(cls, attr_name).items():
                attrs[k] = v
            setattr(cls, attr_name, attrs)

    @classmethod
    def __getsuperattrs(cls):
        superattr = "_supermodels" + cls.__name__
        if hasattr(cls, superattr):
            return
        supermodels = cls.__bases__
        superattr_val = []
        for supermodel in supermodels:
            try:
                supermodel.__getsuperattrs()
                superattr_val.append(supermodel)
            except Exception:
                pass
        setattr(cls, superattr, superattr_val)
        sup = getattr(cls, superattr)
        if not sup:
            return
        cls._merge_inheritable_dicts_(sup)

    def __new__(cls):
        if cls._declared_tablename_ is None:
            cls.tablename = make_tablename(cls.__name__)
        cls.__getsuperattrs()
        return super(Model, cls).__new__(cls)

    def __init__(self):
        if not hasattr(self, 'migrate'):
            self.migrate = self.config.get('migrate', self.db._migrate)
        if not hasattr(self, 'format'):
            self.format = None
        if not hasattr(self, 'primary_keys'):
            self.primary_keys = []
        self._fieldset_pk = set(self.primary_keys or ['id'])

    @property
    def config(self):
        return self.db.config

    def __parse_relation_via(self, via):
        if via is None:
            return via
        rv = sdict()
        splitted = via.split('.')
        rv.via = splitted[0]
        if len(splitted) > 1:
            rv.field = splitted[1]
        return rv

    def __parse_belongs_relation(self, item, on_delete):
        rv = sdict(fk=None, on_delete=on_delete, compound=None)
        if isinstance(item, dict):
            rv.name = list(item)[0]
            rdata = item[rv.name]
            target = None
            if isinstance(rdata, dict):
                if "target" in rdata:
                    target = rdata["target"]
                if "on_delete" in rdata:
                    rv.on_delete = rdata["on_delete"]
            else:
                target = rdata
            if not target:
                target = camelize(rv.name)
            if "." in target:
                target, rv.fk = target.split(".")
            if target == "self":
                target = self.__class__.__name__
            rv.model = target
        else:
            rv.name = item
            rv.model = camelize(item)
        return rv

    def __build_relation_modelname(self, name, relation, singularize):
        relation.model = camelize(name)
        if singularize:
            relation.model = relation.model[:-1]

    def __build_relation_fieldnames(self, relation):
        splitted = relation.model.split('.')
        relation.model = splitted[0]
        if len(splitted) > 1:
            relation.fields = [splitted[1]]
        else:
            if len(self.primary_keys) > 1:
                relation.fields = [
                    f"{decamelize(self.__class__.__name__)}_{pk}"
                    for pk in self.primary_keys
                ]
            else:
                relation.fields = [decamelize(self.__class__.__name__)]

    def __parse_relation_dict(self, rel, singularize):
        if 'scope' in rel.model:
            rel.scope = rel.model['scope']
        if 'where' in rel.model:
            rel.where = rel.model['where']
        if 'via' in rel.model:
            rel.update(self.__parse_relation_via(rel.model['via']))
            del rel.model
        else:
            if 'target' in rel.model:
                rel.model = rel.model['target']
            if not isinstance(rel.model, str):
                self.__build_relation_modelname(rel.name, rel, singularize)

    def __parse_many_relation(self, item, singularize=True):
        rv = ReferenceData(self)
        if isinstance(item, dict):
            rv.name = list(item)[0]
            rv.model = item[rv.name]
            if isinstance(rv.model, dict):
                if 'method' in rv.model:
                    if 'field' in rv.model:
                        rv.fields = [rv.model['field']]
                    else:
                        if len(self.primary_keys) > 1:
                            rv.fields = [
                                f"{decamelize(self.__class__.__name__)}_{pk}"
                                for pk in self.primary_keys
                            ]
                        else:
                            rv.fields = [decamelize(self.__class__.__name__)]
                    rv.cast = rv.model.get('cast')
                    rv.method = rv.model['method']
                    del rv.model
                else:
                    self.__parse_relation_dict(rv, singularize)
        else:
            rv.name = item
            self.__build_relation_modelname(item, rv, singularize)
        if rv.model:
            if not rv.fields:
                self.__build_relation_fieldnames(rv)
            if rv.model == "self":
                rv.model = self.__class__.__name__
        if not rv.via:
            rv.reverse = (
                rv.fields[0] if len(rv.fields) == 1 else
                decamelize(self.__class__.__name__)
            )
        return rv

    def _define_props_(self):
        #: create pydal's Field elements
        self.fields = []
        if not self.primary_keys and 'id' not in self._all_fields_:
            idfield = Field('id')._make_field('id', model=self)
            setattr(self.__class__, 'id', idfield)
            self.fields.append(idfield)
        for name, obj in self._all_fields_.items():
            if obj.modelname is not None:
                obj = Field(obj._type, *obj._args, _kw=obj._ormkw, **obj._kwargs)
                setattr(self.__class__, name, obj)
            self.fields.append(obj._make_field(name, self))

    def __find_matching_fk_definition(self, fields, rmodel):
        match = None
        if not set(fields).issubset(set(rmodel.primary_keys)):
            return match
        for key, val in self.foreign_keys.items():
            if set(val["foreign_fields"]) == set(rmodel.primary_keys):
                match = key
                break
        return match

    def _define_relations_(self):
        self._virtual_relations_ = OrderedDict()
        self._compound_relations_ = {}
        bad_args_error = (
            "belongs_to, has_one and has_many "
            "only accept strings or dicts as arguments"
        )
        #: belongs_to and refers_to are mapped with 'reference' type Field
        _references = []
        _reference_keys = ['_all_belongs_ref_', '_all_refers_ref_']
        belongs_references = {}
        belongs_fks = {}
        for key in _reference_keys:
            if hasattr(self, key):
                _references.append(list(getattr(self, key).values()))
            else:
                _references.append([])
        is_belongs, ondelete = True, 'cascade'
        for _references_obj in _references:
            for item in _references_obj:
                if not isinstance(item, (str, dict)):
                    raise RuntimeError(bad_args_error)
                reference = self.__parse_belongs_relation(item, ondelete)
                reference.is_refers = not is_belongs
                refmodel = (
                    self.db[reference.model]._model_
                    if reference.model != self.__class__.__name__
                    else self
                )
                ref_multi_pk = len(refmodel._fieldset_pk) > 1
                fk_def_key, fks_data, multi_fk = None, {}, []
                if ref_multi_pk and reference.fk:
                    fk_def_key = self.__find_matching_fk_definition(
                        [reference.fk], refmodel
                    )
                    if not fk_def_key:
                        raise SyntaxError(
                            f"{self.__class__.__name__}.{reference.name} relation "
                            "targets a compound primary key table. A matching foreign "
                            "key needs to be defined into `foreign_keys`"
                        )
                    fks_data = self.foreign_keys[fk_def_key]
                elif ref_multi_pk and not reference.fk:
                    multi_fk = list(refmodel.primary_keys)
                elif not reference.fk:
                    reference.fk = list(refmodel._fieldset_pk)[0]
                if multi_fk:
                    references = []
                    fks_data["fields"] = []
                    fks_data["foreign_fields"] = []
                    for fk in multi_fk:
                        refclone = sdict(reference)
                        refclone.fk = fk
                        refclone.ftype = getattr(refmodel, refclone.fk).type
                        refclone.name = f"{refclone.name}_{refclone.fk}"
                        refclone.compound = reference.name
                        references.append(refclone)
                        fks_data["fields"].append(refclone.name)
                        fks_data["foreign_fields"].append(refclone.fk)
                    belongs_fks[reference.name] = sdict(
                        model=reference.model,
                        name=reference.name,
                        local_fields=fks_data["fields"],
                        foreign_fields=fks_data["foreign_fields"],
                        coupled_fields=[
                            (local, fks_data["foreign_fields"][idx])
                            for idx, local in enumerate(fks_data["fields"])
                        ],
                        is_refers=reference.is_refers
                    )
                    self._compound_relations_[reference.name] = sdict(
                        model=reference.model,
                        local_fields=belongs_fks[reference.name].local_fields,
                        foreign_fields=belongs_fks[reference.name].foreign_fields,
                        coupled_fields=belongs_fks[reference.name].coupled_fields,
                    )
                else:
                    reference.ftype = getattr(refmodel, reference.fk).type
                    references = [reference]
                    belongs_fks[reference.name] = sdict(
                        model=reference.model,
                        name=reference.name,
                        local_fields=[reference.name],
                        foreign_fields=[reference.fk],
                        coupled_fields=[(reference.name, reference.fk)],
                        is_refers=reference.is_refers
                    )
                if not fk_def_key and fks_data:
                    self.foreign_keys[reference.name] = self.foreign_keys.get(
                        reference.name
                    ) or fks_data
                for reference in references:
                    if reference.model != self.__class__.__name__:
                        tablename = self.db[reference.model]._tablename
                    else:
                        tablename = self.tablename
                    fieldobj = Field(
                        (
                            f"reference {tablename}" if not ref_multi_pk else
                            f"reference {tablename}.{reference.fk}"
                        ),
                        ondelete=reference.on_delete,
                        _isrefers=not is_belongs
                    )
                    setattr(self.__class__, reference.name, fieldobj)
                    self.fields.append(
                        getattr(self, reference.name)._make_field(
                            reference.name, self
                        )
                    )
                    belongs_references[reference.name] = reference
            is_belongs = False
            ondelete = 'nullify'
        setattr(self.__class__, '_belongs_ref_', belongs_references)
        setattr(self.__class__, '_belongs_fks_', belongs_fks)
        #: has_one are mapped with rowattr
        hasone_references = {}
        if hasattr(self, '_all_hasone_ref_'):
            for item in getattr(self, '_all_hasone_ref_').values():
                if not isinstance(item, (str, dict)):
                    raise RuntimeError(bad_args_error)
                reference = self.__parse_many_relation(item, False)
                if reference.via is not None:
                    #: maps has_one({'thing': {'via': 'otherthings'}})
                    wrapper = HasOneViaWrap
                else:
                    #: maps has_one('thing'), has_one({'thing': 'othername'})
                    wrapper = HasOneWrap
                self._virtual_relations_[reference.name] = rowattr(
                    reference.name
                )(wrapper(reference))
                hasone_references[reference.name] = reference
        setattr(self.__class__, '_hasone_ref_', hasone_references)
        #: has_many are mapped with rowattr
        hasmany_references = {}
        if hasattr(self, '_all_hasmany_ref_'):
            for item in getattr(self, '_all_hasmany_ref_').values():
                if not isinstance(item, (str, dict)):
                    raise RuntimeError(bad_args_error)
                reference = self.__parse_many_relation(item)
                if reference.via is not None:
                    #: maps has_many({'things': {'via': 'otherthings'}})
                    wrapper = HasManyViaWrap
                else:
                    #: maps has_many('things'), has_many({'things': 'othername'})
                    wrapper = HasManyWrap
                self._virtual_relations_[reference.name] = rowattr(
                    reference.name
                )(wrapper(reference))
                hasmany_references[reference.name] = reference
        setattr(self.__class__, '_hasmany_ref_', hasmany_references)
        self.__define_fks()

    def __define_fks(self):
        self._foreign_keys_ = {}
        implicit_defs = {}
        grouped_rels = {}
        for rname, rel in self._belongs_ref_.items():
            rmodel = (
                self.db[rel.model]._model_
                if rel.model != self.__class__.__name__
                else self
            )
            if (
                not rmodel.primary_keys and
                getattr(rmodel, list(rmodel._fieldset_pk)[0]).type == 'id'
            ):
                continue
            if len(rmodel._fieldset_pk) > 1:
                match = self.__find_matching_fk_definition([rel.fk], rmodel)
                if not match:
                    raise SyntaxError(
                        f"{self.__class__.__name__}.{rname} relation targets a "
                        "compound primary key table. A matching foreign key "
                        "needs to be defined into `foreign_keys`."
                    )
                trels = grouped_rels[rmodel.tablename] = grouped_rels.get(
                    rmodel.tablename, {
                        'rels': {},
                        'on_delete': self.foreign_keys[match].get(
                            "on_delete", "cascade"
                        )
                    }
                )
                trels['rels'][rname] = rel
            else:
                # NOTE: we need this since pyDAL doesn't support id/refs types != int
                implicit_defs[rname] = {
                    'table': rmodel.tablename,
                    'fields_local': [rname],
                    'fields_foreign': [rel.fk],
                    'on_delete': Field._internal_delete[rel.on_delete]
                }
        for rname, rel in implicit_defs.items():
            constraint_name =  self.__create_fk_contraint_name(
                rel['table'], *rel['fields_local']
            )
            self._foreign_keys_[constraint_name] = {**rel}
        for tname, rels in grouped_rels.items():
            constraint_name = self.__create_fk_contraint_name(
                tname, *[rel.name for rel in rels['rels'].values()]
            )
            self._foreign_keys_[constraint_name] = {
                'table': tname,
                'fields_local': [rel.name for rel in rels['rels'].values()],
                'fields_foreign': [rel.fk for rel in rels['rels'].values()],
                'on_delete': Field._internal_delete[rels['on_delete']]
            }

    def _define_virtuals_(self):
        self._all_rowattrs_ = {}
        self._all_rowmethods_ = {}
        self._super_rowmethods_ = {}
        err = 'rowattr or rowmethod cannot have the name of an existent field!'
        field_names = [field.name for field in self.fields]
        for attr in ['_virtual_relations_', '_all_virtuals_']:
            for obj in getattr(self, attr, {}).values():
                if obj.field_name in field_names:
                    raise RuntimeError(err)
                wrapped = wrap_virtual_on_model(self, obj.f)
                if isinstance(obj, rowmethod):
                    self._all_rowmethods_[obj.field_name] = wrapped
                    f = Field.Method(obj.field_name, wrapped)
                else:
                    self._all_rowattrs_[obj.field_name] = wrapped
                    f = Field.Virtual(obj.field_name, wrapped)
                self.fields.append(f)
        for obj in self._super_virtuals_.values():
            wrapped = wrap_virtual_on_model(self, obj.f)
            if not isinstance(obj, rowmethod):
                continue
            self._super_rowmethods_[obj.field_name] = wrapped

    def _set_row_persistence_id(self, row, ret):
        row.id = ret.id
        object.__setattr__(row, '_concrete', True)

    def _set_row_persistence_pk(self, row, ret):
        row[self.primary_keys[0]] = ret[self.primary_keys[0]]
        object.__setattr__(row, '_concrete', True)

    def _set_row_persistence_pks(self, row, ret):
        for field_name in self.primary_keys:
            row[field_name] = ret[field_name]
        object.__setattr__(row, '_concrete', True)

    def _unset_row_persistence(self, row):
        for field_name in self._fieldset_pk:
            row[field_name] = None
        object.__setattr__(row, '_concrete', False)

    def _build_rowclass_(self):
        #: build helpers for rows
        save_excluded_fields = (
            set(
                field.name for field in self.fields if
                getattr(field, "type", None) == "id"
            ) |
            set(self._all_rowattrs_.keys()) |
            set(self._all_rowmethods_.keys())
        )
        self._fieldset_initable = set([
            field.name for field in self.fields
        ]) - save_excluded_fields
        self._fieldset_editable = set([
            field.name for field in self.fields
        ]) - save_excluded_fields - self._fieldset_pk
        self._fieldset_all = self._fieldset_initable | self._fieldset_pk
        self._fieldset_update = set([
            field.name for field in self.fields
            if getattr(field, "update", None) is not None
        ]) & self._fieldset_editable
        self._relations_wrapset = (
            set(self._belongs_fks_.keys()) -
            set(self._compound_relations_.keys())
        )
        if not self.primary_keys:
            self._set_row_persistence = self._set_row_persistence_id
        elif len(self.primary_keys) == 1:
            self._set_row_persistence = self._set_row_persistence_pk
        else:
            self._set_row_persistence = self._set_row_persistence_pks
        #: create dynamic row class
        clsname = self.__class__.__name__ + "Row"
        attrs = {'_model': self}
        attrs.update({k: RowFieldMapper(k) for k in self._fieldset_all})
        attrs.update({
            k: RowVirtualMapper(k, v)
            for k, v in self._all_rowattrs_.items()
        })
        attrs.update(self._all_rowmethods_)
        attrs.update({
            k: RowRelationMapper(self.db, self._belongs_ref_[k])
            for k in self._relations_wrapset
        })
        attrs.update({
            k: RowCompoundRelationMapper(
                self.db, data
            ) for k, data in self._compound_relations_.items()
        })
        self._rowclass_ = type(clsname, (StructuredRow,), attrs)
        globals()[clsname] = self._rowclass_

    def _define_(self):
        self.__define_indexes()
        self.__define_validation()
        self.__define_access()
        self.__define_defaults()
        self.__define_updates()
        self.__define_representation()
        self.__define_computations()
        self.__define_callbacks()
        self.__define_scopes()
        self.__define_query_helpers()
        self.__define_form_utils()
        self.setup()

    def __define_validation(self):
        for field in self.fields:
            if isinstance(field, (Field.Method, Field.Virtual)):
                continue
            validation = self.validation.get(field.name, {})
            if isinstance(validation, dict):
                for key in list(validation):
                    field._requires[key] = validation[key]
            elif isinstance(validation, list):
                field._custom_requires += validation
            else:
                field._custom_requires.append(validation)
            field._parse_validation()

    def __define_access(self):
        for field, value in self.fields_rw.items():
            if field == 'id' and field not in self.table:
                continue
            if isinstance(value, (tuple, list)):
                readable, writable = value
            else:
                writable = value
                readable = value
            self.table[field].writable = writable
            self.table[field].readable = readable

    def __define_defaults(self):
        for field, value in self.default_values.items():
            self.table[field].default = value

    def __define_updates(self):
        for field, value in self.update_values.items():
            self.table[field].update = value

    def __define_representation(self):
        for field, value in self.repr_values.items():
            self.table[field].represent = value

    def __define_computations(self):
        err = 'computations should have the name of an existing field to compute!'
        field_names = [field.name for field in self.fields]
        for obj in self._all_computations_.values():
            if obj.field_name not in field_names:
                raise RuntimeError(err)
            self.table[obj.field_name].compute = (
                lambda row, obj=obj, self=self: obj.compute(self, row)
            )

    def __define_callbacks(self):
        for obj in self._all_callbacks_.values():
            for t in obj.t:
                if t in [
                    "_before_insert",
                    "_before_delete",
                    "_after_delete",
                    "_before_save",
                    "_after_save",
                    "_before_destroy",
                    "_after_destroy",
                    "_before_commit_insert",
                    "_before_commit_update",
                    "_before_commit_delete",
                    "_before_commit_save",
                    "_before_commit_destroy",
                    "_after_commit_insert",
                    "_after_commit_update",
                    "_after_commit_delete",
                    "_after_commit_save",
                    "_after_commit_destroy"
                ]:
                    getattr(self.table, t).append(
                        lambda a, obj=obj, self=self: obj.f(self, a)
                    )
                else:
                    getattr(self.table, t).append(
                        lambda a, b, obj=obj, self=self: obj.f(self, a, b)
                    )

    def __define_scopes(self):
        self._scopes_ = {}
        for obj in self._all_scopes_.values():
            self._scopes_[obj.name] = obj
            if not hasattr(self.__class__, obj.name):
                setattr(
                    self.__class__, obj.name,
                    classmethod(wrap_scope_on_model(obj.f))
                )

    def __prepend_table_name(self, name, ns):
        return '%s_%s__%s' % (self.tablename, ns, name)

    def __create_index_name(self, *values):
        components = []
        for value in values:
            components.append(value.replace('_', ''))
        return self.__prepend_table_name("_".join(components), 'widx')

    def __create_fk_contraint_name(self, *values):
        components = []
        for value in values:
            components.append(value.replace('_', ''))
        return self.__prepend_table_name("fk__" + "_".join(components), 'ecnt')

    def __parse_index_dict(self, value):
        rv = {}
        fields = value.get('fields') or []
        if not isinstance(fields, (list, tuple)):
            fields = [fields]
        rv['fields'] = fields
        where_query = None
        where_cond = value.get('where')
        if callable(where_cond):
            where_query = where_cond(self.__class__)
        if where_query:
            rv['where'] = where_query
        expressions = []
        expressions_cond = value.get('expressions')
        if callable(expressions_cond):
            expressions = expressions_cond(self.__class__)
        if not isinstance(expressions, (tuple, list)):
            expressions = [expressions]
        rv['expressions'] = expressions
        rv['unique'] = value.get('unique', False)
        return rv

    def __define_indexes(self):
        self._indexes_ = {}
        #: auto-define indexes based on fields
        for field in self.fields:
            if getattr(field, 'unique', False):
                idx_name = self.__prepend_table_name(f'{field.name}_unique', 'widx')
                idx_dict = self.__parse_index_dict(
                    {'fields': [field.name], 'unique': True}
                )
                self._indexes_[idx_name] = idx_dict
        #: parse user-defined fields
        for key, value in self.indexes.items():
            if isinstance(value, bool):
                if not value:
                    continue
                if not isinstance(key, tuple):
                    key = [key]
                if any(field not in self.table for field in key):
                    raise SyntaxError(f'Invalid field specified in indexes: {key}')
                idx_name = self.__create_index_name(*key)
                idx_dict = {'fields': key, 'expressions': [], 'unique': False}
            elif isinstance(value, dict):
                idx_name = self.__prepend_table_name(key, 'widx')
                idx_dict = self.__parse_index_dict(value)
            else:
                raise SyntaxError('Values in indexes dict should be booleans or dicts')
            self._indexes_[idx_name] = idx_dict

    def _row_record_query_id(self, row):
        return self.table.id == row.id

    def _row_record_query_pk(self, row):
        return self.table[self.primary_keys[0]] == row[self.primary_keys[0]]

    def _row_record_query_pks(self, row):
        return reduce(
            operator.and_, [self.table[pk] == row[pk] for pk in self.primary_keys]
        )

    def __define_query_helpers(self):
        if not self.primary_keys:
            self._query_id = self.table.id != None # noqa
            self._query_row = self._row_record_query_id
            self._order_by_id_asc = self.table.id
            self._order_by_id_desc = ~self.table.id
        elif len(self.primary_keys) == 1:
            self._query_id = self.table[self.primary_keys[0]] != None # noqa
            self._query_row = self._row_record_query_pk
            self._order_by_id_asc = self.table[self.primary_keys[0]]
            self._order_by_id_desc = ~self.table[self.primary_keys[0]]
        else:
            self._query_id = reduce(
                operator.and_, [
                    self.table[key] != None # noqa
                    for key in self.primary_keys
                ]
            )
            self._query_row = self._row_record_query_pks
            self._order_by_id_asc = reduce(
                operator.or_, [self.table[key] for key in self.primary_keys]
            )
            self._order_by_id_desc = reduce(
                operator.or_, [~self.table[key] for key in self.primary_keys]
            )

    def __define_form_utils(self):
        #: labels
        for field, value in self.form_labels.items():
            self.table[field].label = value
        #: info
        for field, value in self.form_info.items():
            self.table[field].comment = value
        #: widgets
        for field, value in self.form_widgets.items():
            self.table[field].widget = value

    def setup(self):
        pass

    def get_rowmethod(self, name: str):
        return self._all_rowmethods_[name]

    def super_rowmethod(self, name: str):
        return self._super_rowmethods_[name]

    @classmethod
    def _instance_(cls):
        return cls.table._model_

    @classmethod
    def new(cls, **attributes):
        inst = cls._instance_()
        attrset = set(attributes.keys())
        rowattrs = {}
        for field in (inst._fieldset_initable - inst._relations_wrapset) & attrset:
            rowattrs[field] = attributes[field]
        for field in inst._fieldset_initable - attrset:
            val = cls.table[field].default
            if callable(val):
                val = val()
            rowattrs[field] = val
        for field in (inst.primary_keys or ["id"]):
            if inst.table[field].type == "id":
                rowattrs[field] = None
        for field in set(inst._compound_relations_.keys()) & attrset:
            reldata = inst._compound_relations_[field]
            for local_field, foreign_field in reldata.coupled_fields:
                rowattrs[local_field] = attributes[field][foreign_field]
        rv = inst._rowclass_(
            rowattrs, __concrete=False,
            **{k: attributes[k] for k in attrset - set(rowattrs)}
        )
        rv._fields.update({
            field: attributes[field] if not attributes[field] else (
                typed_row_reference_from_record(
                    attributes[field], inst.db[inst._belongs_fks_[field].model]._model_
                ) if isinstance(attributes[field], StructuredRow) else
                typed_row_reference(
                    attributes[field], inst.db[inst._belongs_fks_[field].model]
                )
            ) for field in inst._relations_wrapset & attrset
        })
        return rv

    @classmethod
    def create(cls, *args, skip_callbacks=False, **kwargs):
        inst = cls._instance_()
        if args:
            if isinstance(args[0], (dict, sdict)):
                for key in list(args[0]):
                    kwargs[key] = args[0][key]
        for field in set(inst._compound_relations_.keys()) & set(kwargs.keys()):
            reldata = inst._compound_relations_[field]
            for local_field, foreign_field in reldata.coupled_fields:
                kwargs[local_field] = kwargs[field][foreign_field]
        return cls.table.validate_and_insert(skip_callbacks=skip_callbacks, **kwargs)

    @classmethod
    def validate(cls, row, write_values: bool = False):
        inst, errors = cls._instance_(), sdict()
        for field_name in inst._fieldset_all:
            field = inst.table[field_name]
            default = getattr(field, 'default')
            if callable(default):
                default = default()
            value = row.get(field_name, default)
            new_value, error = field.validate(value)
            if error:
                errors[field_name] = error
            elif new_value is not None and write_values:
                row[field_name] = new_value
        return errors

    @classmethod
    def where(cls, cond):
        if not isinstance(cond, types.LambdaType):
            raise ValueError("Model.where expects a lambda as parameter")
        return cls.db.where(cond(cls), model=cls)

    @classmethod
    def all(cls):
        return cls.db.where(cls._instance_()._query_id, model=cls)

    @classmethod
    def first(cls):
        return cls.all().select(
            orderby=cls._instance_()._order_by_id_asc,
            limitby=(0, 1)
        ).first()

    @classmethod
    def last(cls):
        return cls.all().select(
            orderby=cls._instance_()._order_by_id_desc,
            limitby=(0, 1)
        ).first()

    @classmethod
    def get(cls, *args, **kwargs):
        if args:
            inst = cls._instance_()
            if len(args) == 1:
                if isinstance(args[0], tuple):
                    args = args[0]
                elif isinstance(args[0], dict) and not kwargs:
                    return cls.table(**args[0])
            if len(args) != len(inst._fieldset_pk):
                raise SyntaxError(
                    f"{cls.__name__}.get requires the same number of arguments "
                    "as its primary keys"
                )
            pks = inst.primary_keys or ["id"]
            return cls.table(
                **{pks[idx]: val for idx, val in enumerate(args)}
            )
        return cls.table(**kwargs)

    @rowmethod('update_record')
    def _update_record(self, row, skip_callbacks=False, **fields):
        newfields = fields or dict(row)
        for field_name in set(newfields.keys()) - self._fieldset_editable:
            del newfields[field_name]
        res = self.db(
            self._query_row(row), ignore_common_filters=True
        ).update(skip_callbacks=skip_callbacks, **newfields)
        if res:
            row.update(self.get(**{key: row[key] for key in self._fieldset_pk}))
        return row

    @rowmethod('delete_record')
    def _delete_record(self, row, skip_callbacks=False):
        return self.db(self._query_row(row)).delete(skip_callbacks=skip_callbacks)

    @rowmethod('refresh')
    def _row_refresh(self, row) -> bool:
        if not row._concrete:
            return False
        last = self.db(self._query_row(row)).select(
            limitby=(0, 1),
            orderby_on_limitby=False
        ).first()
        if not last:
            return False
        row._fields.update(last._fields)
        row.__dict__.clear()
        row._changes.clear()
        return True

    @rowmethod('save')
    def _row_save(
        self,
        row,
        raise_on_error: bool = False,
        skip_callbacks: bool = False
    ) -> bool:
        if row._concrete:
            if set(row._changes.keys()) & self._fieldset_pk:
                if raise_on_error:
                    raise SaveException(
                        'Cannot save a record with altered primary key(s)'
                    )
                return False
            for field_name in self._fieldset_update:
                val = self.table[field_name].update
                if callable(val):
                    val = val()
                row[field_name] = val
        errors = self.validate(row, write_values=True)
        if errors:
            if raise_on_error:
                raise ValidationError
            return False
        if row._concrete:
            res = self.db(
                self._query_row(row), ignore_common_filters=True
            )._update_from_save(self, row, skip_callbacks=skip_callbacks)
            if not res:
                if raise_on_error:
                    raise UpdateFailureOnSave
                return False
        else:
            self.table._insert_from_save(row, skip_callbacks=skip_callbacks)
            if not row._concrete:
                if raise_on_error:
                    raise InsertFailureOnSave
                return False
        extra_changes = {
            key: row._changes[key]
            for key in set(row._changes.keys()) & set(row.__dict__.keys())
        }
        row._changes.clear()
        row._changes.update(extra_changes)
        return True

    @rowmethod('destroy')
    def _row_destroy(
        self,
        row,
        raise_on_error: bool = False,
        skip_callbacks: bool = False
    ) -> bool:
        if not row._concrete:
            return False
        res = self.db(
            self._query_row(row), ignore_common_filters=True
        )._delete_from_destroy(self, row, skip_callbacks=skip_callbacks)
        if not res:
            if raise_on_error:
                raise DestroyException
            return False
        row._changes.clear()
        return True


class RowFieldMapper:
    __slots__ = ["field"]

    def __init__(self, field: str):
        self.field = field

    def __set__(self, obj, val):
        obj._fields[self.field] = val

    def __get__(self, obj, objtype=None):
        return obj._fields.get(self.field)

    def __delete__(self, obj):
        obj._fields.pop(self.field, None)


class RowVirtualMapper:
    __slots__ = ["field", "fget"]

    def __init__(self, field: str, fget: Callable[..., Any]):
        self.field = field
        self.fget = fget

    def __get__(self, obj, objtype=None):
        if self.field not in obj._virtuals:
            obj._virtuals[self.field] = rv = self.fget(obj)
            return rv
        return obj._virtuals[self.field]

    def __delete__(self, obj):
        obj._virtuals.pop(self.field, None)


class RowRelationMapper:
    __slots__ = ["table", "field"]

    def __init__(self, db, relation_data):
        self.table = db[relation_data.model]
        self.field = relation_data.name

    def __set__(self, obj, val):
        if not val:
            val = None
        elif val and not isinstance(val, RowReferenceMixin):
            if isinstance(val, StructuredRow):
                val = typed_row_reference_from_record(val, val._model)
            else:
                val = typed_row_reference(val, self.table)
        obj._fields[self.field] = val

    def __get__(self, obj, objtype=None):
        return obj._fields.get(self.field)

    def __delete__(self, obj):
        obj._fields.pop(self.field, None)


class RowCompoundRelationMapper:
    __slots__ = ["table", "fields", "field"]

    def __init__(self, db, relation_data):
        self.table = db[relation_data.model]
        self.field = relation_data.name
        self.fields = relation_data.coupled_fields

    def __set__(self, obj, val):
        if not isinstance(val, (StructuredRow, RowReferenceMulti)):
            return
        for local_field, foreign_field in self.fields:
            obj[local_field] = val[foreign_field]

    def __get__(self, obj, objtype=None):
        pks = {fk: obj[lk] for lk, fk in self.fields}
        key = (self.field, *pks.values())
        if key not in obj._compound_rels:
            obj._compound_rels[key] = RowReferenceMulti(pks, self.table) if all(
                v is not None for v in pks.values()
            ) else None
        return obj._compound_rels[key]

    def __delete__(self, obj):
        pks = {fk: obj[lk] for lk, fk in self.fields}
        key = (self.field, *pks.values())
        obj._compound_rels.pop(key, None)



# -*- coding: utf-8 -*-
"""
    tests.session
    -------------

    Test Emmett session module
"""

import pytest

from emmett.asgi.wrappers import Request
from emmett.ctx import RequestContext, current
from emmett.sessions import SessionManager
from emmett.testing.env import ScopeBuilder
from emmett.wrappers.response import Response


class FakeRequestContext(RequestContext):
    def __init__(self, app, scope):
        self.request = Request(scope, None, None)
        self.response = Response()
        self.session = None


@pytest.fixture(scope='module')
def ctx():
    builder = ScopeBuilder()
    token = current._init_(FakeRequestContext(None, builder.get_data()[0]))
    yield current
    current._close_(token)


@pytest.mark.asyncio
async def test_session_cookie(ctx):
    session_cookie = SessionManager.cookies(
        key='sid',
        secure=True,
        domain='localhost',
        cookie_name='foo_session'
    )
    assert session_cookie.key == 'sid'
    assert session_cookie.secure is True
    assert session_cookie.domain == 'localhost'

    await session_cookie.open_request()
    assert ctx.session._expiration == 3600

    await session_cookie.close_request()
    cookie = str(ctx.response.cookies)
    assert 'foo_session' in cookie
    assert 'Domain=localhost;' in cookie
    assert 'secure' in cookie.lower()

    ctx.request.cookies = ctx.response.cookies
    await session_cookie.open_request()
    assert ctx.session._expiration == 3600



# -*- coding: utf-8 -*-
"""
    emmett.orm.migrations.base
    --------------------------

    Provides base migrations objects.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Dict, Type

from ...datastructures import sdict
from .. import Database, Model, Field
from .engine import MetaEngine, Engine
from .helpers import WrappedOperation, _feasible_as_dbms_default

if TYPE_CHECKING:
    from .operations import Operation


class Schema(Model):
    tablename = "emmett_schema"
    version = Field()


class Migration:
    _registered_ops_: Dict[str, Type[Operation]] = {}
    skip_on_compare: bool = False

    @classmethod
    def register_operation(
        cls,
        name: str
    ) -> Callable[[Type[Operation]], Type[Operation]]:
        def wrap(op_cls: Type[Operation]) -> Type[Operation]:
            cls._registered_ops_[name] = op_cls
            return op_cls
        return wrap

    def __init__(self, app: Any, db: Database, is_meta: bool = False):
        self.db = db
        if is_meta:
            self.engine = MetaEngine(db)
        else:
            self.engine = Engine(db)

    def __getattr__(self, name: str) -> WrappedOperation:
        registered = self._registered_ops_.get(name)
        if registered is not None:
            return WrappedOperation(registered, name, self.engine)
        raise NotImplementedError


class Column(sdict):
    def __init__(
        self,
        name: str,
        type: str = 'string',
        unique: bool = False,
        notnull: bool = False,
        **kwargs: Any
    ):
        self.name = name
        self.type = type
        self.unique = unique
        self.notnull = notnull
        for key, val in kwargs.items():
            self[key] = val
        self.length: int = self.length or 255

    def _fk_type(self, db: Database, tablename: str):
        if self.name not in db[tablename]._model_._belongs_ref_:
            return
        ref = db[tablename]._model_._belongs_ref_[self.name]
        if ref.ftype != 'id':
            self.type = ref.ftype
            self.length = db[ref.model][ref.fk].length
            self.on_delete = None

    @classmethod
    def from_field(cls, field: Field) -> Column:
        rv = cls(
            field.name,
            field._pydal_types.get(field._type, field._type),
            field.unique,
            field.notnull,
            length=field.length,
            ondelete=field.ondelete,
            **field._ormkw
        )
        if _feasible_as_dbms_default(field.default):
            rv.default = field.default
        rv._fk_type(field.db, field.tablename)
        return rv

    def __repr__(self) -> str:
        return "%s(%s)" % (
            self.__class__.__name__,
            ", ".join(["%s=%r" % (k, v) for k, v in self.items()])
        )



# -*- coding: utf-8 -*-
"""
    emmett.server
    -------------

    Provides server wrapper over granian

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from typing import Optional

from granian import Granian


def run(
    interface,
    app,
    host='127.0.0.1',
    port=8000,
    loop='auto',
    loop_opt=False,
    log_level=None,
    workers=1,
    threads=1,
    threading_mode='workers',
    backlog=1024,
    enable_websockets=True,
    ssl_certfile: Optional[str] = None,
    ssl_keyfile: Optional[str] = None
):
    app_path = ":".join([app[0], app[1] or "app"])
    runner = Granian(
        app_path,
        address=host,
        port=port,
        interface=interface,
        workers=workers,
        threads=threads,
        pthreads=threads,
        threading_mode=threading_mode,
        loop=loop,
        loop_opt=loop_opt,
        websockets=enable_websockets,
        backlog=backlog,
        log_level=log_level,
        ssl_cert=ssl_certfile,
        ssl_key=ssl_keyfile
    )
    runner.serve()



# -*- coding: utf-8 -*-
"""
    emmett.wrappers.request
    -----------------------

    Provides http request wrappers.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from abc import abstractmethod
from cgi import FieldStorage, parse_header
from io import BytesIO
from urllib.parse import parse_qs
from typing import Any

import pendulum

from ..datastructures import sdict
from ..parsers import Parsers
from ..utils import cachedprop
from . import IngressWrapper
from .helpers import FileStorage


class Request(IngressWrapper):
    __slots__ = ['_now', 'method']

    method: str

    @property
    @abstractmethod
    async def body(self) -> bytes: ...

    @cachedprop
    def now(self) -> pendulum.DateTime:
        return pendulum.instance(self._now)

    @cachedprop
    def now_local(self) -> pendulum.DateTime:
        return self.now.in_timezone(pendulum.local_timezone())  # type: ignore

    @cachedprop
    def content_type(self) -> str:
        return parse_header(self.headers.get('content-type', ''))[0]

    @cachedprop
    def content_length(self) -> int:
        return self.headers.get('content_length', 0, cast=int)

    _empty_body_methods = {v: v for v in ['GET', 'HEAD', 'OPTIONS']}

    @cachedprop
    async def _input_params(self):
        if self._empty_body_methods.get(self.method) or not self.content_type:
            return sdict(), sdict()
        return await self._load_params()

    @cachedprop
    async def body_params(self) -> sdict[str, Any]:
        rv, _ = await self._input_params
        return rv

    @cachedprop
    async def files(self) -> sdict[str, FileStorage]:
        _, rv = await self._input_params
        return rv

    def _load_params_missing(self, data):
        return sdict(), sdict()

    def _load_params_json(self, data):
        try:
            params = Parsers.get_for('json')(data)
        except Exception:
            params = {}
        return sdict(params), sdict()

    def _load_params_form_urlencoded(self, data):
        rv = sdict()
        for key, values in parse_qs(
            data.decode('latin-1'), keep_blank_values=True
        ).items():
            if len(values) == 1:
                rv[key] = values[0]
                continue
            rv[key] = values
        return rv, sdict()

    @property
    def _multipart_headers(self):
        return self.headers

    @staticmethod
    def _file_param_from_field(field):
        return FileStorage(
            BytesIO(field.file.read()),
            field.filename,
            field.name,
            field.type,
            field.headers
        )

    def _load_params_form_multipart(self, data):
        params, files = sdict(), sdict()
        field_storage = FieldStorage(
            BytesIO(data),
            headers=self._multipart_headers,
            environ={'REQUEST_METHOD': self.method},
            keep_blank_values=True
        )
        for key in field_storage:
            field = field_storage[key]
            if isinstance(field, list):
                if len(field) > 1:
                    pvalues, fvalues = [], []
                    for item in field:
                        if item.filename is not None:
                            fvalues.append(self._file_param_from_field(item))
                        else:
                            pvalues.append(item.value)
                    if pvalues:
                        params[key] = pvalues
                    if fvalues:
                        files[key] = fvalues
                    continue
                else:
                   field = field[0]
            if field.filename is not None:
                files[key] = self._file_param_from_field(field)
            else:
                params[key] = field.value
        return params, files

    _params_loaders = {
        'application/json': _load_params_json,
        'application/x-www-form-urlencoded': _load_params_form_urlencoded,
        'multipart/form-data': _load_params_form_multipart
    }

    async def _load_params(self):
        loader = self._params_loaders.get(
            self.content_type, self._load_params_missing)
        return loader(self, await self.body)

    @abstractmethod
    async def push_promise(self, path: str): ...



# -*- coding: utf-8 -*-
"""
    emmett.orm.objects
    ------------------

    Provides pyDAL objects implementation for Emmett.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import copy
import datetime
import decimal
import operator
import types

from collections import OrderedDict, defaultdict
from enum import Enum
from functools import reduce
from typing import Any, Dict, Optional

from pydal.objects import (
    Table as _Table,
    Field as _Field,
    Set as _Set,
    Row as _Row,
    Rows as _Rows,
    IterRows as _IterRows,
    Query,
    Expression
)

from ..ctx import current
from ..datastructures import sdict
from ..html import tag
from ..serializers import xml_encode
from ..utils import cachedprop
from ..validators import ValidateFromDict
from .helpers import (
    RelationBuilder,
    GeoFieldWrapper,
    RowReferenceMixin,
    wrap_scope_on_set,
    typed_row_reference_from_record
)

type_int = int


class Table(_Table):
    def __init__(self, db, tablename, *fields, **kwargs):
        _primary_keys, _notnulls = list(kwargs.get('primarykey', [])), {}
        _notnulls = {
            field.name: field.notnull
            for field in fields if hasattr(field, 'notnull')
        }
        super(Table, self).__init__(db, tablename, *fields, **kwargs)
        self._before_save = []
        self._after_save = []
        self._before_destroy = []
        self._after_destroy = []
        self._before_commit = []
        self._before_commit_insert = []
        self._before_commit_update = []
        self._before_commit_delete = []
        self._before_commit_save = []
        self._before_commit_destroy = []
        self._after_commit = []
        self._after_commit_insert = []
        self._after_commit_update = []
        self._after_commit_delete = []
        self._after_commit_save = []
        self._after_commit_destroy = []
        self._unique_fields_validation_ = {}
        self._primary_keys = _primary_keys
        #: avoid pyDAL mess in ops and migrations
        if len(self._primary_keys) == 1 and getattr(self, '_primarykey', None):
            del self._primarykey
        for key in self._primary_keys:
            self[key].notnull = _notnulls[key]
        if not hasattr(self, '_id'):
            self._id = None

    @cachedprop
    def _has_commit_insert_callbacks(self):
        return any([
            self._before_commit,
            self._after_commit,
            self._before_commit_insert,
            self._after_commit_insert
        ])

    @cachedprop
    def _has_commit_update_callbacks(self):
        return any([
            self._before_commit,
            self._after_commit,
            self._before_commit_update,
            self._after_commit_update
        ])

    @cachedprop
    def _has_commit_delete_callbacks(self):
        return any([
            self._before_commit,
            self._after_commit,
            self._before_commit_delete,
            self._after_commit_delete
        ])

    @cachedprop
    def _has_commit_save_callbacks(self):
        return any([
            self._before_commit,
            self._after_commit,
            self._before_commit_save,
            self._after_commit_save
        ])

    @cachedprop
    def _has_commit_destroy_callbacks(self):
        return any([
            self._before_commit,
            self._after_commit,
            self._before_commit_destroy,
            self._after_commit_destroy
        ])

    def _create_references(self):
        self._referenced_by = []
        self._referenced_by_list = []
        self._references = []

    def _fields_and_values_for_save(self, row, fieldset, op_method):
        fields = {key: row[key] for key in fieldset}
        return op_method(fields)

    def insert(self, skip_callbacks=False, **fields):
        row = self._fields_and_values_for_insert(fields)
        if not skip_callbacks and any(f(row) for f in self._before_insert):
            return 0
        ret = self._db._adapter.insert(self, row.op_values())
        if not skip_callbacks:
            if self._has_commit_insert_callbacks:
                txn = self._db._adapter.top_transaction()
                if txn:
                    txn._add_op(TransactionOp(
                        TransactionOps.insert,
                        self,
                        TransactionOpContext(
                            values=row,
                            ret=ret
                        )
                    ))
            if ret:
                for f in self._after_insert:
                    f(row, ret)
        return ret

    def validate_and_insert(self, skip_callbacks=False, **fields):
        response, new_fields = self._validate_fields(fields)
        if not response.errors:
            response.id = self.insert(skip_callbacks=skip_callbacks, **new_fields)
        return response

    def _insert_from_save(self, row, skip_callbacks=False):
        if not skip_callbacks and any(f(row) for f in self._before_save):
            return row
        fields = self._fields_and_values_for_save(
            row, self._model_._fieldset_initable, self._fields_and_values_for_insert
        )
        ret = self.insert(skip_callbacks=skip_callbacks, **fields)
        if ret:
            self._model_._set_row_persistence(row, ret)
        if not skip_callbacks:
            if self._has_commit_save_callbacks:
                txn = self._db._adapter.top_transaction()
                if txn:
                    txn._add_op(TransactionOp(
                        TransactionOps.save,
                        self,
                        TransactionOpContext(
                            values=fields,
                            ret=ret,
                            row=row.clone_changed(),
                            changes=row.changes
                        )
                    ))
            if row._concrete:
                for f in self._after_save:
                    f(row)
        return row


class Field(_Field):
    _internal_types = {
        'integer': 'int',
        'double': 'float',
        'boolean': 'bool',
        'list:integer': 'list:int'
    }
    _pydal_types = {
        'int': 'integer',
        'bool': 'boolean',
        'list:int': 'list:integer',
    }
    _internal_delete = {
        'cascade': 'CASCADE', 'nullify': 'SET NULL', 'nothing': 'NO ACTION'
    }
    _inst_count_ = 0
    _obj_created_ = False

    def __init__(self, type='string', *args, **kwargs):
        self.modelname = None
        #: convert type
        self._type = self._internal_types.get(type, type)
        #: convert 'rw' -> 'readable', 'writeable'
        if 'rw' in kwargs:
            _rw = kwargs.pop('rw')
            if isinstance(_rw, (tuple, list)):
                read, write = _rw
            else:
                read = write = _rw
            kwargs['readable'] = read
            kwargs['writable'] = write
        #: convert 'info' -> 'comment'
        _info = kwargs.pop('info', None)
        if _info:
            kwargs['comment'] = _info
        #: convert ondelete parameter
        _ondelete = kwargs.get('ondelete')
        if _ondelete:
            if _ondelete not in list(self._internal_delete):
                raise SyntaxError(
                    'Field ondelete should be set on %s, %s or %s' %
                    list(self._internal_delete)
                )
            kwargs['ondelete'] = self._internal_delete[_ondelete]
        #: process 'refers_to' fields
        self._isrefers = kwargs.pop('_isrefers', None)
        #: get auto validation preferences
        self._auto_validation = kwargs.pop('auto_validation', True)
        #: intercept validation (will be processed by `_make_field`)
        self._requires = {}
        self._custom_requires = []
        if 'validation' in kwargs:
            _validation = kwargs.pop('validation')
            if isinstance(_validation, dict):
                self._requires = _validation
            else:
                self._custom_requires = _validation
                if not isinstance(self._custom_requires, list):
                    self._custom_requires = [self._custom_requires]
        self._validation = {}
        self._vparser = ValidateFromDict()
        #: ensure 'length' is an integer
        if 'length' in kwargs:
            kwargs['length'] = int(kwargs['length'])
        #: store args and kwargs for `_make_field`
        self._ormkw = kwargs.pop('_kw', {})
        self._args = args
        self._kwargs = kwargs
        #: increase creation counter (used to keep order of fields)
        self._inst_count_ = Field._inst_count_
        Field._inst_count_ += 1

    def _default_validation(self):
        rv = {}
        auto_types = [
            'int', 'float', 'date', 'time', 'datetime', 'json'
        ]
        if self._type in auto_types:
            rv['is'] = self._type
        elif self._type.startswith('decimal'):
            rv['is'] = 'decimal'
        elif self._type == 'jsonb':
            rv['is'] = 'json'
        if self._type == 'bigint':
            rv['is'] = 'int'
        if self._type == 'bool':
            rv['in'] = (False, True)
        if self._type in ['string', 'text', 'password']:
            rv['len'] = {'lte': self.length}
        if self._type == 'password':
            rv['len']['gte'] = 6
            rv['crypt'] = True
        if self._type == 'list:int':
            rv['is'] = 'list:int'
        if (
            self.notnull or self._type.startswith('reference') or
            self._type.startswith('list:reference')
        ):
            rv['presence'] = True
        if not self.notnull and self._isrefers is True:
            rv['allow'] = 'empty'
        if self.unique:
            rv['unique'] = True
        return rv

    def _parse_validation(self):
        for key in list(self._requires):
            self._validation[key] = self._requires[key]
        self.requires = self._vparser(self, self._validation) + \
            self._custom_requires

    #: `_make_field` will be called by `Model` class or `Form` class
    #  it will make internal Field class compatible with the pyDAL's one
    def _make_field(self, name, model=None):
        if self._obj_created_:
            return self
        if model is not None:
            self.modelname = model.__class__.__name__
        #: convert field type to pyDAL ones if needed
        ftype = self._pydal_types.get(self._type, self._type)
        if ftype.startswith("geo") and model:
            geometry_type = self._ormkw["geometry_type"]
            srid = self._ormkw["srid"] or getattr(model.db._adapter, "srid", 4326)
            dimension = self._ormkw["dimension"] or 2
            ftype = f"{ftype}({geometry_type},{srid},{dimension})"
        #: create pyDAL's Field instance
        super(Field, self).__init__(name, ftype, *self._args, **self._kwargs)
        #: add automatic validation (if requested)
        if self._auto_validation:
            auto = True
            if self.modelname:
                auto = model.auto_validation
            if auto:
                self._validation = self._default_validation()
        #: validators
        if not self.modelname:
            self._parse_validation()
        self._obj_created_ = True
        return self

    def __str__(self):
        if self._obj_created_:
            return super(Field, self).__str__()
        return object.__str__(self)

    def __repr__(self):
        if self.modelname and hasattr(self, 'name'):
            return "<%s.%s (%s) field>" % (self.modelname, self.name,
                                           self._type)
        return super(Field, self).__repr__()

    @classmethod
    def string(cls, *args, **kwargs):
        return cls('string', *args, **kwargs)

    @classmethod
    def int(cls, *args, **kwargs):
        return cls('int', *args, **kwargs)

    @classmethod
    def bigint(cls, *args, **kwargs):
        return cls('bigint', *args, **kwargs)

    @classmethod
    def float(cls, *args, **kwargs):
        return cls('float', *args, **kwargs)

    @classmethod
    def text(cls, *args, **kwargs):
        return cls('text', *args, **kwargs)

    @classmethod
    def bool(cls, *args, **kwargs):
        return cls('bool', *args, **kwargs)

    @classmethod
    def blob(cls, *args, **kwargs):
        return cls('blob', *args, **kwargs)

    @classmethod
    def date(cls, *args, **kwargs):
        return cls('date', *args, **kwargs)

    @classmethod
    def time(cls, *args, **kwargs):
        return cls('time', *args, **kwargs)

    @classmethod
    def datetime(cls, *args, **kwargs):
        return cls('datetime', *args, **kwargs)

    @classmethod
    def decimal(cls, precision, scale, *args, **kwargs):
        return cls('decimal({},{})'.format(precision, scale), *args, **kwargs)

    @classmethod
    def json(cls, *args, **kwargs):
        return cls('json', *args, **kwargs)

    @classmethod
    def jsonb(cls, *args, **kwargs):
        return cls('jsonb', *args, **kwargs)

    @classmethod
    def password(cls, *args, **kwargs):
        return cls('password', *args, **kwargs)

    @classmethod
    def upload(cls, *args, **kwargs):
        return cls('upload', *args, **kwargs)

    @classmethod
    def int_list(cls, *args, **kwargs):
        return cls('list:int', *args, **kwargs)

    @classmethod
    def string_list(cls, *args, **kwargs):
        return cls('list:string', *args, **kwargs)

    @classmethod
    def geography(
        cls,
        geometry_type: str = 'GEOMETRY',
        srid: Optional[type_int] = None,
        dimension: Optional[type_int] = None,
        **kwargs
    ):
        kwargs['_kw'] = {
            "geometry_type": geometry_type,
            "srid": srid,
            "dimension": dimension
        }
        return cls("geography", **kwargs)

    @classmethod
    def geometry(
        cls,
        geometry_type: str = 'GEOMETRY',
        srid: Optional[type_int] = None,
        dimension: Optional[type_int] = None,
        **kwargs
    ):
        kwargs['_kw'] = {
            "geometry_type": geometry_type,
            "srid": srid,
            "dimension": dimension
        }
        return cls("geometry", **kwargs)

    def cast(self, value, **kwargs):
        return Expression(
            self.db, self._dialect.cast, self,
            self._dialect.types[value] % kwargs, value)


class Set(_Set):
    def __init__(self, db, query, ignore_common_filters=None, model=None):
        super(Set, self).__init__(db, query, ignore_common_filters)
        self._model_ = model
        self._scopes_ = {}
        self._load_scopes_()

    def _load_scopes_(self):
        if self._model_ is None:
            tablemap = self.db._adapter.tables(self.query)
            if len(tablemap) == 1:
                self._model_ = tablemap.popitem()[1]._model_
        if self._model_:
            self._scopes_ = self._model_._instance_()._scopes_

    def _clone(self, ignore_common_filters=None, model=None, **changes):
        return self.__class__(
            self.db, changes.get('query', self.query),
            ignore_common_filters=ignore_common_filters,
            model=model or self._model_
        )

    def where(self, query, ignore_common_filters=None, model=None):
        if query is None:
            return self
        elif isinstance(query, Table):
            query = self.db._adapter.id_query(query)
        elif isinstance(query, str):
            query = Expression(self.db, query)
        elif isinstance(query, Field):
            query = query != None
        elif isinstance(query, types.LambdaType):
            model = model or self._model_
            if not model:
                raise ValueError(
                    "Too many models involved in the Set to use a lambda")
            query = query(model)
        q = self.query & query if self.query else query
        return self._clone(ignore_common_filters, model, query=q)

    def _parse_paginate(self, pagination):
        if isinstance(pagination, tuple):
            offset = pagination[0]
            limit = pagination[1]
        else:
            offset = pagination
            limit = 10
        return ((offset - 1) * limit, offset * limit)

    def _join_set_builder(self, obj, jdata, auto_select_tables):
        return JoinedSet._from_set(
            obj, jdata=jdata, auto_select_tables=auto_select_tables
        )

    def _left_join_set_builder(self, jdata):
        return JoinedSet._from_set(
            self, ljdata=jdata, auto_select_tables=[self._model_.table]
        )

    def _run_select_(self, *fields, **options):
        tablemap = self.db._adapter.tables(
            self.query,
            options.get('join', None),
            options.get('left', None),
            options.get('orderby', None),
            options.get('groupby', None)
        )
        fields, concrete_tables = self.db._adapter._expand_all_with_concrete_tables(
            fields, tablemap
        )
        options['_concrete_tables'] = concrete_tables
        return self.db._adapter.select(self.query, fields, options)

    def _get_table_from_query(self) -> Table:
        if self._model_:
            return self._model_.table
        return self.db._adapter.get_table(self.query)

    def select(self, *fields, **options):
        obj = self
        pagination, including = (
            options.pop('paginate', None),
            options.pop('including', None)
        )
        if pagination:
            options['limitby'] = self._parse_paginate(pagination)
        if including and self._model_ is not None:
            options['left'], jdata = self._parse_left_rjoins(including)
            obj = self._left_join_set_builder(jdata)
        return obj._run_select_(*fields, **options)

    def iterselect(self, *fields, **options):
        pagination = options.pop('paginate', None)
        if pagination:
            options['limitby'] = self._parse_paginate(pagination)
        tablemap = self.db._adapter.tables(
            self.query,
            options.get('join', None),
            options.get('left', None),
            options.get('orderby', None),
            options.get('groupby', None)
        )
        fields, concrete_tables = self.db._adapter._expand_all_with_concrete_tables(
            fields, tablemap
        )
        options['_concrete_tables'] = concrete_tables
        return self.db._adapter.iterselect(self.query, fields, options)

    def update(self, skip_callbacks=False, **update_fields):
        table = self._get_table_from_query()
        row = table._fields_and_values_for_update(update_fields)
        if not row._values:
            raise ValueError("No fields to update")
        if not skip_callbacks and any(f(self, row) for f in table._before_update):
            return 0
        ret = self.db._adapter.update(table, self.query, row.op_values())
        if not skip_callbacks:
            if table._has_commit_update_callbacks:
                txn = self._db._adapter.top_transaction()
                if txn:
                    txn._add_op(TransactionOp(
                        TransactionOps.update,
                        table,
                        TransactionOpContext(
                            values=row,
                            dbset=self,
                            ret=ret
                        )
                    ))
            ret and [f(self, row) for f in table._after_update]
        return ret

    def delete(self, skip_callbacks=False):
        table = self._get_table_from_query()
        if not skip_callbacks and any(f(self) for f in table._before_delete):
            return 0
        ret = self.db._adapter.delete(table, self.query)
        if not skip_callbacks:
            if table._has_commit_delete_callbacks:
                txn = self._db._adapter.top_transaction()
                if txn:
                    txn._add_op(TransactionOp(
                        TransactionOps.delete,
                        table,
                        TransactionOpContext(
                            dbset=self,
                            ret=ret
                        )
                    ))
            ret and [f(self) for f in table._after_delete]
        return ret

    def validate_and_update(self, skip_callbacks=False, **update_fields):
        table = self._get_table_from_query()
        current._dbvalidation_record_id_ = None
        if table._unique_fields_validation_ and self.count() == 1:
            if any(
                table._unique_fields_validation_.get(fieldname)
                for fieldname in update_fields.keys()
            ):
                current._dbvalidation_record_id_ = \
                    self.select(table.id).first().id
        response = Row()
        response.errors = Row()
        new_fields = copy.copy(update_fields)
        for key, value in update_fields.items():
            value, error = table[key].validate(value)
            if error:
                response.errors[key] = '%s' % error
            else:
                new_fields[key] = value
        del current._dbvalidation_record_id_
        if response.errors:
            response.updated = None
        else:
            row = table._fields_and_values_for_update(new_fields)
            if not row._values:
                raise ValueError("No fields to update")
            if not skip_callbacks and any(f(self, row) for f in table._before_update):
                ret = 0
            else:
                ret = self.db._adapter.update(
                    table, self.query, row.op_values()
                )
                if not skip_callbacks and ret:
                    for f in table._after_update:
                        f(self, row)
            response.updated = ret
        return response

    def _update_from_save(self, model, row, skip_callbacks=False):
        table: Table = model.table
        if not skip_callbacks and any(f(row) for f in table._before_save):
            return False
        fields = table._fields_and_values_for_save(
            row, model._fieldset_editable, table._fields_and_values_for_update
        )
        if not skip_callbacks and any(f(self, fields) for f in table._before_update):
            return False
        ret = self.db._adapter.update(table, self.query, fields.op_values())
        if not skip_callbacks:
            if table._has_commit_update_callbacks or table._has_commit_save_callbacks:
                txn = self._db._adapter.top_transaction()
                if txn and table._has_commit_update_callbacks:
                    txn._add_op(TransactionOp(
                        TransactionOps.update,
                        table,
                        TransactionOpContext(
                            values=fields,
                            dbset=self,
                            ret=ret
                        )
                    ))
                if txn and table._has_commit_save_callbacks:
                    txn._add_op(TransactionOp(
                        TransactionOps.save,
                        table,
                        TransactionOpContext(
                            values=fields,
                            dbset=self,
                            ret=ret,
                            row=row.clone_changed(),
                            changes=row.changes
                        )
                    ))
            ret and [f(self, fields) for f in table._after_update]
            ret and [f(row) for f in table._after_save]
        return bool(ret)

    def _delete_from_destroy(self, model, row, skip_callbacks=False):
        table: Table = model.table
        if not skip_callbacks and any(f(row) for f in table._before_destroy):
            return False
        if not skip_callbacks and any(f(self) for f in table._before_delete):
            return 0
        ret = self.db._adapter.delete(table, self.query)
        if ret:
            model._unset_row_persistence(row)
        if not skip_callbacks:
            if (
                table._has_commit_delete_callbacks or
                table._has_commit_destroy_callbacks
            ):
                txn = self._db._adapter.top_transaction()
                if txn and table._has_commit_delete_callbacks:
                    txn._add_op(TransactionOp(
                        TransactionOps.delete,
                        table,
                        TransactionOpContext(
                            dbset=self,
                            ret=ret
                        )
                    ))
                if txn and table._has_commit_destroy_callbacks:
                    txn._add_op(TransactionOp(
                        TransactionOps.destroy,
                        table,
                        TransactionOpContext(
                            dbset=self,
                            ret=ret,
                            row=row.clone_changed(),
                            changes=row.changes
                        )
                    ))
            ret and [f(self) for f in table._after_delete]
            ret and [f(row) for f in table._after_destroy]
        return bool(ret)

    def join(self, *args):
        rv = self
        if self._model_ is not None:
            joins = []
            jdata = []
            auto_select_tables = [self._model_.table]
            for arg in args:
                condition, table, rel_type = self._parse_rjoin(arg)
                joins.append(condition)
                jdata.append((arg, table._tablename, rel_type))
                auto_select_tables.append(table)
            if joins:
                q = joins[0]
                for join in joins[1:]:
                    q = q & join
                rv = rv.where(q, model=self._model_)
                return self._join_set_builder(rv, jdata, auto_select_tables)
        return rv

    def switch(self, model):
        self._model_ = model
        return self

    def _parse_rjoin(self, arg):
        #: match has_many
        rel = self._model_._hasmany_ref_.get(arg)
        if rel:
            if rel.via:
                r = RelationBuilder(rel, self._model_._instance_()).via()
                return r[0], r[1]._table, 'many'
            r = RelationBuilder(rel, self._model_._instance_())
            return r.many(), rel.table, 'many'
        #: match belongs_to and refers_to
        rel = self._model_._belongs_fks_.get(arg)
        if rel:
            r = RelationBuilder(rel, self._model_._instance_()).belongs_query()
            return r, self._model_.db[rel.model], 'belongs'
        #: match has_one
        rel = self._model_._hasone_ref_.get(arg)
        if rel:
            if rel.via:
                r = RelationBuilder(rel, self._model_._instance_()).via()
                return r[0], r[1]._table, 'one'
            r = RelationBuilder(rel, self._model_._instance_())
            return r.many(), rel.table, 'one'
        raise RuntimeError(
            f'Unable to find {arg} relation of {self._model_.__name__} model'
        )

    def _parse_left_rjoins(self, args):
        if not isinstance(args, (list, tuple)):
            args = [args]
        joins = []
        jdata = []
        for arg in args:
            condition, table, rel_type = self._parse_rjoin(arg)
            joins.append(table.on(condition))
            jdata.append((arg, table._tablename, rel_type))
        return joins, jdata

    def _jcolnames_from_rowstmps(self, tmps):
        colnames = []
        all_colnames = {}
        jcolnames = {}
        for colname in tmps:
            all_colnames[colname[0]] = colname[0]
            jcolnames[colname[0]] = jcolnames.get(colname[0], [])
            jcolnames[colname[0]].append(colname[1])
        for colname in all_colnames.keys():
            if colname == self._stable_:
                colnames.append(colname)
                del jcolnames[colname]
        return colnames, jcolnames

    def __getattr__(self, name):
        scope = self._scopes_.get(name)
        if scope:
            return wrap_scope_on_set(self, self._model_._instance_(), scope.f)
        raise AttributeError(name)


class RelationSet(object):
    _relation_method_ = 'many'

    def __init__(self, db, relation_builder, row):
        self.db = db
        self._relation_ = relation_builder
        self._row_ = row

    def _get_query_(self):
        try:
            return getattr(self._relation_, self._relation_method_)(self._row_)
        except AttributeError as e:
            raise RuntimeError(e.message)
        except Exception:
            raise

    @property
    def _model_(self):
        return self._relation_.ref.model_instance

    @cachedprop
    def _fields_(self):
        pks = self._relation_.model.primary_keys or ["id"]
        return [
            (relation_field.name, pks[idx])
            for idx, relation_field in enumerate(self._relation_.ref.fields_instances)
        ]

    @cachedprop
    def _scopes_(self):
        if self._relation_.ref.method:
            return [lambda: self._relation_.ref.dbset.query]
        return self._relation_._extra_scopes(self._relation_.ref)

    @cachedprop
    def _set(self):
        return Set(self.db, self._get_query_(), model=self._model_.__class__)

    def __getattr__(self, name):
        return getattr(self._set, name)

    def __repr__(self):
        return repr(self._set)

    def __call__(self, query, ignore_common_filters=False):
        return self._set.where(query, ignore_common_filters)

    def _last_resultset(self, refresh=False):
        if refresh or not hasattr(self, '_cached_resultset'):
            self._cached_resultset = self._cache_resultset()
        return self._cached_resultset

    def _filter_reload(self, kwargs):
        return kwargs.pop('reload', False)

    def new(self, **kwargs):
        attrs = self._get_fields_from_scopes(self._scopes_, self._model_.tablename)
        attrs.update(**kwargs)
        for ref, local in self._fields_:
            attrs[ref] = self._row_[local]
        return self._model_.new(**attrs)

    def create(self, skip_callbacks=False, **kwargs):
        attrs = self._get_fields_from_scopes(self._scopes_, self._model_.tablename)
        attrs.update(**kwargs)
        for ref, local in self._fields_:
            attrs[ref] = self._row_[local]
        return self._model_.create(skip_callbacks=skip_callbacks, **attrs)

    @staticmethod
    def _get_fields_from_scopes(scopes, table_name):
        rv = {}
        for scope in scopes:
            query = scope()
            components = [query.second, query.first]
            current_kv = []
            while components:
                component = components.pop()
                if isinstance(component, Query):
                    components.append(component.second)
                    components.append(component.first)
                else:
                    if (
                        isinstance(component, Field) and
                        component._tablename == table_name
                    ):
                        current_kv.append(component)
                    else:
                        if current_kv:
                            current_kv.append(component)
                        else:
                            components.pop()
                if len(current_kv) > 1:
                    rv[current_kv[0].name] = current_kv[1]
                    current_kv = []
        return rv


class HasOneSet(RelationSet):
    def _cache_resultset(self):
        return self.select(self._model_.table.ALL, limitby=(0, 1)).first()

    def __call__(self, *args, **kwargs):
        refresh = self._filter_reload(kwargs)
        if not args and not kwargs:
            return self._last_resultset(refresh)
        kwargs['limitby'] = (0, 1)
        return self.select(*args, **kwargs).first()


class HasManySet(RelationSet):
    def _cache_resultset(self):
        return self.select(self._model_.table.ALL)

    def __call__(self, *args, **kwargs):
        refresh = self._filter_reload(kwargs)
        if not args and not kwargs:
            return self._last_resultset(refresh)
        return self.select(*args, **kwargs)

    def add(self, obj, skip_callbacks=False):
        if not isinstance(obj, (StructuredRow, RowReferenceMixin)):
            raise RuntimeError(f"Unsupported parameter {obj}")
        attrs = self._get_fields_from_scopes(
            self._scopes_, self._model_.tablename
        )
        rev_attrs = {}
        for ref, local in self._fields_:
            attrs[ref] = self._row_[local]
            rev_attrs[local] = attrs[ref]
        rv = self.db(self._model_._query_row(obj)).validate_and_update(
            skip_callbacks=skip_callbacks, **attrs
        )
        if rv:
            for key, val in attrs.items():
                obj[key] = val
            if len(rev_attrs) > 1:
                comprel_key = (self._relation_.ref.reverse, *rev_attrs.values())
                obj._compound_rels[comprel_key] = typed_row_reference_from_record(
                    self._row_, self._row_._model
                )
        return rv

    def remove(self, obj, skip_callbacks=False):
        if not isinstance(obj, (StructuredRow, RowReferenceMixin)):
            raise RuntimeError(f"Unsupported parameter {obj}")
        attrs, is_delete = {ref: None for ref, _ in self._fields_}, False
        if self._model_._belongs_fks_[self._relation_.ref.reverse].is_refers:
            rv = self.db(self._model_._query_row(obj)).validate_and_update(
                skip_callbacks=skip_callbacks,
                **attrs
            )
        else:
            is_delete = True
            rv = self.db(self._model_._query_row(obj)).delete(
                skip_callbacks=skip_callbacks
            )
        if rv:
            for key, val in attrs.items():
                obj[key] = val
            if len(attrs) > 1:
                obj._compound_rels[(self._relation_.ref.reverse, None, None)] = None
            if is_delete:
                obj._concrete = False
        return rv


class ViaSet(RelationSet):
    _relation_method_ = 'via'

    @cachedprop
    def _viadata(self):
        query, rfield, model_name, rid, via, viadata = super()._get_query_()
        return sdict(
            query=query,
            rfield=rfield,
            model_name=model_name,
            rid=rid,
            via=via,
            data=viadata
        )

    def _get_query_(self):
        return self._viadata.query

    @property
    def _model_(self):
        return self.db[self._viadata.model_name]._model_


class HasOneViaSet(ViaSet):
    def _cache_resultset(self):
        return self.select(self._viadata.rfield, limitby=(0, 1)).first()

    def __call__(self, *args, **kwargs):
        refresh = self._filter_reload(kwargs)
        if not args:
            if not kwargs:
                return self._last_resultset(refresh)
            args = [self._viadata.rfield]
        kwargs['limitby'] = (0, 1)
        return self.select(*args, **kwargs).first()

    def create(self, **kwargs):
        raise RuntimeError('Cannot create third objects for one via relations')


class HasManyViaSet(ViaSet):
    _via_error = "Can't %s elements in has_many relations without a join table"

    def _cache_resultset(self):
        return self.select(self._viadata.rfield)

    def __call__(self, *args, **kwargs):
        refresh = self._filter_reload(kwargs)
        if not args:
            if not kwargs:
                return self._last_resultset(refresh)
            args = [self._viadata.rfield]
        return self.select(*args, **kwargs)

    def _get_relation_fields(self):
        viadata = self._viadata.data
        manyref = self._model_._hasmany_ref_[viadata.via]
        fkdata = manyref.model_instance._belongs_fks_
        jref = fkdata[viadata.field or viadata.name[:-1]]
        fields_src = manyref.fields
        fields_dst = list(jref.coupled_fields)
        return fields_src, fields_dst

    def _fields_from_scopes(self):
        viadata = self._viadata.data
        rel = self._model_._hasmany_ref_[viadata.via]
        if rel.method:
            scopes = [lambda: rel.dbset.query]
        else:
            scopes = self._relation_._extra_scopes(rel)
        return self._get_fields_from_scopes(scopes, rel.table_name)

    def create(self, **kwargs):
        raise RuntimeError('Cannot create third objects for many via relations')

    def add(self, obj, skip_callbacks=False, **kwargs):
        # works on join tables only!
        if self._viadata.via is None:
            raise RuntimeError(self._via_error % 'add')
        nrow = self._fields_from_scopes()
        nrow.update(**kwargs)
        #: get belongs references
        self_fields, rel_fields = self._get_relation_fields()
        for idx, self_field in enumerate(self_fields):
            nrow[self_field] = self._viadata.rid[idx]
        for local_field, foreign_field in rel_fields:
            nrow[local_field] = obj[foreign_field]
        #: validate and insert
        return self.db[self._viadata.via]._model_.create(
            nrow, skip_callbacks=skip_callbacks
        )

    def remove(self, obj, skip_callbacks=False):
        # works on join tables only!
        if self._viadata.via is None:
            raise RuntimeError(self._via_error % 'remove')
        #: get belongs references
        self_fields, rel_fields = self._get_relation_fields()
        #: delete
        query = reduce(
            operator.and_, [
                self.db[self._viadata.via][field] == self._viadata.rid[idx]
                for idx, field in enumerate(self_fields)
            ] + [
                self.db[self._viadata.via][local_field] == obj[foreign_field]
                for local_field, foreign_field in rel_fields
            ]
        )
        return self.db(query).delete(skip_callbacks=skip_callbacks)


class JoinedSet(Set):
    @classmethod
    def _from_set(cls, obj, jdata=[], ljdata=[], auto_select_tables=[]):
        rv = cls(obj.db, obj.query, obj.query.ignore_common_filters, obj._model_)
        rv._stable_ = obj._model_.tablename
        rv._jdata_ = list(jdata)
        rv._ljdata_ = list(ljdata)
        rv._auto_select_tables_ = list(auto_select_tables)
        rv._pks_ = obj._model_._instance_()._fieldset_pk
        return rv

    def _clone(self, ignore_common_filters=None, model=None, **changes):
        rv = super()._clone(ignore_common_filters, model, **changes)
        rv._stable_ = self._stable_
        rv._jdata_ = self._jdata_
        rv._ljdata_ = self._ljdata_
        rv._auto_select_tables_ = self._auto_select_tables_
        rv._pks_ = self._pks_
        return rv

    def _join_set_builder(self, obj, jdata, auto_select_tables):
        return JoinedSet._from_set(
            obj, jdata=self._jdata_ + jdata, ljdata=self._ljdata_,
            auto_select_tables=self._auto_select_tables_ + auto_select_tables
        )

    def _left_join_set_builder(self, jdata):
        return JoinedSet._from_set(
            self, jdata=self._jdata_, ljdata=self._ljdata_ + jdata,
            auto_select_tables=self._auto_select_tables_
        )

    def _iterselect_rows(self, *fields, **options):
        tablemap = self.db._adapter.tables(
            self.query,
            options.get('join', None),
            options.get('left', None),
            options.get('orderby', None),
            options.get('groupby', None)
        )
        fields, concrete_tables = self.db._adapter._expand_all_with_concrete_tables(
            fields, tablemap
        )
        colnames, sql = self.db._adapter._select_wcols(self.query, fields, **options)
        return JoinIterRows(self.db, sql, fields, concrete_tables, colnames)

    def _split_joins(self, joins):
        rv = {'belongs': [], 'one': [], 'many': []}
        for jname, jtable, rel_type in joins:
            rv[rel_type].append((jname, jtable))
        return rv['belongs'], rv['one'], rv['many']

    def _build_records_from_joined(self, rowmap, inclusions, colnames):
        for rid, many_data in inclusions.items():
            for jname, included in many_data.items():
                rowmap[rid][jname]._cached_resultset = Rows(
                    self.db, list(included.values()), []
                )
        return JoinRows(
            self.db, list(rowmap.values()), colnames,
            _jdata=self._jdata_ + self._ljdata_
        )

    def _select_rowpks_extractor(self, row):
        if not set(row.keys()).issuperset(self._pks_):
            return None
        if len(self._pks_) > 1:
            return tuple(row[pk] for pk in self._pks_)
        return row[tuple(self._pks_)[0]]

    def _run_select_(self, *fields, **options):
        #: build parsers
        belongs_j, one_j, many_j = self._split_joins(self._jdata_)
        belongs_l, one_l, many_l = self._split_joins(self._ljdata_)
        parsers = (
            self._build_jparsers(belongs_j, one_j, many_j) +
            self._build_lparsers(belongs_l, one_l, many_l)
        )
        #: auto add selection field for left joins
        if self._ljdata_:
            fields = list(fields)
            if not fields:
                fields = [v.ALL for v in self._auto_select_tables_]
            for join in options['left']:
                fields.append(join.first.ALL)
        #: use iterselect for performance
        rows = self._iterselect_rows(*fields, **options)
        #: rebuild rowset using nested objects
        plainrows = []
        rowmap = OrderedDict()
        inclusions = defaultdict(
            lambda: {
                jname: OrderedDict() for jname, _ in (many_j + many_l)
            }
        )
        for row in rows:
            if self._stable_ not in row:
                plainrows.append(row)
                continue
            rid = self._select_rowpks_extractor(row[self._stable_])
            if rid is None:
                plainrows.append(row)
                continue
            rowmap[rid] = rowmap.get(rid, row[self._stable_])
            for parser in parsers:
                parser(rowmap, inclusions, row, rid)
        if not rowmap and plainrows:
            return Rows(self.db, plainrows, rows.colnames)
        return self._build_records_from_joined(
            rowmap, inclusions, rows.colnames
        )

    def _build_jparsers(self, belongs, one, many):
        rv = []
        for jname, jtable in belongs:
            rv.append(self._jbelong_parser(self.db, jname, jtable))
        for jname, jtable in one:
            rv.append(self._jone_parser(self.db, jname, jtable))
        for jname, jtable in many:
            rv.append(self._jmany_parser(self.db, jname, jtable))
        return rv

    def _build_lparsers(self, belongs, one, many):
        rv = []
        for jname, jtable in belongs:
            rv.append(self._lbelong_parser(self.db, jname, jtable))
        for jname, jtable in one:
            rv.append(self._lone_parser(self.db, jname, jtable))
        for jname, jtable in many:
            rv.append(self._lmany_parser(self.db, jname, jtable))
        return rv

    @staticmethod
    def _jbelong_parser(db, fieldname, tablename):
        rmodel = db[tablename]._model_

        def parser(rowmap, inclusions, row, rid):
            rowmap[rid][fieldname] = typed_row_reference_from_record(
                row[tablename], rmodel
            )
        return parser

    @staticmethod
    def _jone_parser(db, fieldname, tablename):
        def parser(rowmap, inclusions, row, rid):
            rowmap[rid][fieldname]._cached_resultset = row[tablename]
        return parser

    @staticmethod
    def _jmany_parser(db, fieldname, tablename):
        rmodel = db[tablename]._model_
        pks = rmodel.primary_keys or ["id"]
        ext = lambda row: tuple(row[pk] for pk in pks) if len(pks) > 1 else row[pks[0]]

        def parser(rowmap, inclusions, row, rid):
            inclusions[rid][fieldname][ext(row[tablename])] = \
                inclusions[rid][fieldname].get(
                    ext(row[tablename]), row[tablename]
                )
        return parser

    @staticmethod
    def _lbelong_parser(db, fieldname, tablename):
        rmodel = db[tablename]._model_
        pks = rmodel.primary_keys or ["id"]
        check = lambda row: all(row[pk] for pk in pks)

        def parser(rowmap, inclusions, row, rid):
            if not check(row[tablename]):
                return
            rowmap[rid][fieldname] = typed_row_reference_from_record(
                row[tablename], rmodel
            )
        return parser

    @staticmethod
    def _lone_parser(db, fieldname, tablename):
        rmodel = db[tablename]._model_
        pks = rmodel.primary_keys or ["id"]
        check = lambda row: all(row[pk] for pk in pks)

        def parser(rowmap, inclusions, row, rid):
            if not check(row[tablename]):
                return
            rowmap[rid][fieldname]._cached_resultset = row[tablename]
        return parser

    @staticmethod
    def _lmany_parser(db, fieldname, tablename):
        rmodel = db[tablename]._model_
        pks = rmodel.primary_keys or ["id"]
        ext = lambda row: tuple(row[pk] for pk in pks) if len(pks) > 1 else row[pks[0]]
        check = lambda row: all(row[pk] for pk in pks)

        def parser(rowmap, inclusions, row, rid):
            if not check(row[tablename]):
                return
            inclusions[rid][fieldname][ext(row[tablename])] = \
                inclusions[rid][fieldname].get(
                    ext(row[tablename]), row[tablename]
                )
        return parser


class Row(_Row):
    _as_dict_types_ = tuple(
        [type(None)] + [int, float, bool, list, dict, str] +
        [datetime.datetime, datetime.date, datetime.time]
    )

    @classmethod
    def _from_engine(cls, data):
        rv = cls()
        rv.__dict__.__init__(**data)
        return rv

    def as_dict(self, datetime_to_str=False, custom_types=None, geo_coordinates=True):
        rv = {}
        for key, val in self.items():
            if isinstance(val, Row):
                val = val.as_dict()
            elif isinstance(val, decimal.Decimal):
                val = float(val)
            elif isinstance(val, GeoFieldWrapper) and geo_coordinates:
                val = val.__json__()
            elif not isinstance(val, self._as_dict_types_):
                continue
            rv[key] = val
        return rv

    def __getstate__(self):
        return self.as_dict(geo_coordinates=False)

    def __json__(self):
        return self.as_dict()

    def __xml__(self, key=None, quote=True):
        return xml_encode(self.as_dict(), key or 'row', quote)

    def __str__(self):
        return '<Row {}>'.format(self.as_dict(geo_coordinates=False))

    def __repr__(self):
        return str(self)

    def __getitem__(self, name):
        try:
            return super(Row, self).__getitem__(name)
        except KeyError:
            raise KeyError(name)

    def __getattr__(self, name):
        try:
            return self.__getitem__(name)
        except KeyError:
            raise AttributeError(name)

    def __eq__(self, other):
        if not isinstance(other, Row):
            return False
        try:
            return self.__dict__ == other.__dict__
        except AttributeError:
            return False

    def __copy__(self):
        return Row(self)


class StructuredRow(Row):
    __slots__ = ["_changes", "_compound_rels", "_concrete", "_fields", "_virtuals"]

    @classmethod
    def _from_engine(cls, data: Dict[str, Any]):
        rv = cls(__concrete=True)
        rv._fields.update(data)
        return rv

    def __init__(
        self,
        fields: Optional[Dict[str, Any]] = None,
        **extras: Any
    ):
        object.__setattr__(self, "_changes", {})
        object.__setattr__(self, "_compound_rels", {})
        object.__setattr__(self, "_concrete", extras.pop("__concrete", False))
        object.__setattr__(self, "_fields", fields or {})
        object.__setattr__(self, "_virtuals", {})
        self.__dict__.__init__(**extras)

    def __contains__(self, name):
        return name in self._fields or name in self.__dict__

    def __getitem__(self, name):
        return object.__getattribute__(self, name)

    def __setattr__(self, key, value):
        if key in self.__slots__:
            return
        oldv = (
            self._changes[key][0] if key in self._changes else
            getattr(self, key, None)
        )
        object.__setattr__(self, key, value)
        newv = getattr(self, key, None)
        if (oldv is None and value is not None) or oldv != newv:
            self._changes[key] = (oldv, newv)
        else:
            self._changes.pop(key, None)

    def __setitem__(self, key, value):
        self.__setattr__(key, value)

    def __getstate__(self):
        return {
            "__fields": self._fields,
            "__extras": self.__dict__,
            "__struct": {
                "_concrete": self._concrete,
                "_changes": {},
                "_compound_rels": {},
                "_virtuals": {}
            }
        }

    def __setstate__(self, state):
        self.__dict__.update(state["__extras"])
        object.__setattr__(self, "_fields", state["__fields"])
        for key, val in state["__struct"].items():
            object.__setattr__(self, key, val)

    def __bool__(self):
        return bool(self._fields) or bool(self.__dict__)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self._fields == other._fields and self.__dict__ == other.__dict__

    def __copy__(self):
        return StructuredRow(self._fields, __concrete=self._concrete, **self.__dict__)

    def keys(self):
        for pool in (self._fields, self.__dict__):
            for item in pool.keys():
                yield item

    def values(self):
        for pool in (self._fields, self.__dict__):
            for item in pool.values():
                yield item

    def items(self):
        for pool in (self._fields, self.__dict__):
            for item in pool.items():
                yield item

    def update(self, *args, **kwargs):
        for arg in args:
            for key, val in arg.items():
                self.__setattr__(key, val)
        for key, val in kwargs.items():
            self.__setattr__(key, val)

    @property
    def changes(self):
        return sdict(self._changes)

    @property
    def has_changed(self):
        return bool(self._changes)

    def has_changed_value(self, key):
        return key in self._changes

    def get_value_change(self, key):
        return self._changes.get(key, None)

    def clone(self):
        fields = {}
        fieldset = set(self._fields.keys())
        changeset = set(self._changes.keys())
        for key in fieldset & changeset:
            fields[key] = self._changes[key][0]
        for key in fieldset - changeset:
            fields[key] = self._fields[key]
        return self.__class__(fields, __concrete=self._concrete, **self.__dict__)

    def clone_changed(self):
        return self.__class__(
            {**self._fields}, __concrete=self._concrete, **self.__dict__
        )

    @property
    def validation_errors(self):
        return self._model.validate(self)

    @property
    def is_valid(self):
        return not bool(self._model.validate(self))


class Rows(_Rows):
    def __init__(
        self, db=None, records=[], colnames=[], compact=True, rawrows=None
    ):
        self.db = db
        self.records = records
        self.colnames = colnames
        self._rowkeys_ = list(self.records[0].keys()) if self.records else []
        self._getrow = self._getrow_compact_ if self.compact else self._getrow_

    @cachedprop
    def compact(self):
        if not self.records:
            return False
        return len(self._rowkeys_) == 1 and self._rowkeys_[0] != '_extra'

    @cachedprop
    def compact_tablename(self):
        if not self._rowkeys_:
            return None
        return self._rowkeys_[0]

    def _getrow_(self, i):
        return self.records[i]

    def _getrow_compact_(self, i):
        return self._getrow_(i)[self.compact_tablename]

    def __getitem__(self, i):
        if isinstance(i, slice):
            return self.__class__(self.db, self.records[i], self.colnames)
        return self._getrow(i)

    def column(self, column=None):
        colname = str(column) if column else self.colnames[0]
        return [r[colname] for r in self]

    def sorted(self, f, reverse=False):
        if self.compact:
            keyf = lambda r: f(r[self.compact_tablename])
        else:
            keyf = f
        return [r for r in sorted(self.records, key=keyf, reverse=reverse)]

    def sort(self, f, reverse=False):
        self.records = self.sorted(f, reverse)

    def append(self, obj):
        row = Row({self.compact_tablename: obj}) if self.compact else obj
        self.records.append(row)

    def insert(self, position, obj):
        row = Row({self.compact_tablename: obj}) if self.compact else obj
        self.records.insert(position, row)

    def render(self, *args, **kwargs):
        raise NotImplementedError

    def as_list(self, datetime_to_str=False, custom_types=None):
        return [item.as_dict(datetime_to_str, custom_types) for item in self]

    def as_dict(self, key='id', datetime_to_str=False, custom_types=None):
        if '.' in key:
            splitted_key = key.split('.')
            keyf = lambda row: row[splitted_key[0]][splitted_key[1]]
        else:
            keyf = lambda row: row[key]
        return {keyf(item): item for item in self.as_list()}

    def __json__(self):
        return [item.__json__() for item in self]

    def __xml__(self, key=None, quote=True):
        key = key or 'rows'
        return tag[key](*[item.__xml__(quote=quote) for item in self])

    def __str__(self):
        return str(self.records)


class IterRows(_IterRows):
    def __init__(self, db, sql, fields, concrete_tables, colnames):
        self.db = db
        self.fields = fields
        self.concrete_tables = concrete_tables
        self.colnames = colnames
        self.fdata, self.tables = self.db._adapter._parse_expand_colnames(fields)
        self.cursor = self.db._adapter.cursor
        self.db._adapter.execute(sql)
        self.db._adapter.lock_cursor(self.cursor)
        self._head = None
        self.last_item = None
        self.last_item_id = None
        self.compact = True
        self.blob_decode = True
        self.cacheable = False
        self.sql = sql

    def __next__(self):
        db_row = self.cursor.fetchone()
        if db_row is None:
            raise StopIteration
        row = self.db._adapter._parse(
            db_row,
            self.fdata,
            self.tables,
            self.concrete_tables,
            self.fields,
            self.colnames,
            self.blob_decode
        )
        if self.compact:
            keys = list(row.keys())
            if len(keys) == 1 and keys[0] != '_extra':
                row = row[keys[0]]
        return row

    def __iter__(self):
        if self._head:
            yield self._head
        try:
            row = next(self)
            while row is not None:
                yield row
                row = next(self)
        except StopIteration:
            self.db._adapter.close_cursor(self.cursor)
        return


class JoinRows(Rows):
    def __init__(self, *args, **kwargs):
        self._joins_ = kwargs.pop('_jdata')
        super(JoinRows, self).__init__(*args, **kwargs)

    def as_list(
        self, compact=True, storage_to_dict=True, datetime_to_str=False,
        custom_types=None
    ):
        if storage_to_dict:
            items = []
            for row in self:
                item = row.as_dict(datetime_to_str, custom_types)
                for jdata in self._joins_:
                    if not jdata[2]:
                        item[jdata[0]] = row[jdata[0]].as_list()
                items.append(item)
        else:
            items = [item for item in self]
        return items


class JoinIterRows(IterRows):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.compact = False


class TransactionOps(str, Enum):
    __str__ = lambda v: v.value

    insert = "insert"
    update = "update"
    delete = "delete"
    save = "save"
    destroy = "destroy"


class TransactionOpContext:
    __slots__ = ["values", "dbset", "return_value", "row", "changes"]

    def __init__(
        self,
        values: Any = None,
        dbset: Optional[Set] = None,
        ret: Any = None,
        row: Optional[Row] = None,
        changes: Optional[sdict] = None
    ):
        self.values = values
        self.dbset = dbset
        self.return_value = ret
        self.row = row
        self.changes = changes


class TransactionOp:
    __slots__ = ["op_type", "table", "context"]

    def __init__(
        self,
        op_type: TransactionOps,
        table: Table,
        context: TransactionOpContext
    ):
        self.op_type = op_type
        self.table = table
        self.context = context



# -*- coding: utf-8 -*-
"""
    emmett.orm.adapters
    -------------------

    Provides ORM adapters facilities.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import sys

from functools import wraps

from pydal.adapters.base import SQLAdapter
from pydal.adapters.mssql import (
    MSSQL1,
    MSSQL3,
    MSSQL4,
    MSSQL1N,
    MSSQL3N,
    MSSQL4N
)
from pydal.adapters.postgres import (
    Postgre,
    PostgrePsyco,
    PostgrePG8000,
    PostgreNew,
    PostgrePsycoNew,
    PostgrePG8000New
)
from pydal.helpers.classes import SQLALL
from pydal.helpers.regex import REGEX_TABLE_DOT_FIELD
from pydal.parsers import ParserMethodWrapper, for_type as _parser_for_type
from pydal.representers import TReprMethodWrapper, for_type as _representer_for_type

from .engines import adapters
from .helpers import GeoFieldWrapper, PasswordFieldWrapper, typed_row_reference
from .objects import Expression, Field, Row, IterRows


adapters._registry_.update({
    'mssql': MSSQL4,
    'mssql2': MSSQL1,
    'mssql3': MSSQL3,
    'mssqln': MSSQL4N,
    'mssqln2': MSSQL1N,
    'mssqln3': MSSQL3N,
    'postgres2': PostgreNew,
    'postgres2:psycopg2': PostgrePsycoNew,
    'postgres2:pg8000': PostgrePG8000New,
    'postgres3': Postgre,
    'postgres3:psycopg2': PostgrePsyco,
    'postgres3:pg8000': PostgrePG8000
})


def _wrap_on_obj(f, adapter):
    @wraps(f)
    def wrapped(*args, **kwargs):
        return f(adapter, *args, **kwargs)
    return wrapped


def patch_adapter(adapter):
    #: BaseAdapter interfaces
    adapter._expand_all_with_concrete_tables = _wrap_on_obj(
        _expand_all_with_concrete_tables, adapter
    )
    adapter._parse = _wrap_on_obj(_parse, adapter)
    adapter._parse_expand_colnames = _wrap_on_obj(_parse_expand_colnames, adapter)
    adapter.iterparse = _wrap_on_obj(iterparse, adapter)
    adapter.parse = _wrap_on_obj(parse, adapter)
    patch_parser(adapter.dialect, adapter.parser)
    patch_representer(adapter.representer)

    #: SQLAdapter interfaces
    if not isinstance(adapter, SQLAdapter):
        return
    adapter._select_aux = _wrap_on_obj(_select_aux, adapter)
    adapter._select_wcols_inner = adapter._select_wcols
    adapter._select_wcols = _wrap_on_obj(_select_wcols, adapter)
    adapter.insert = _wrap_on_obj(insert, adapter)
    adapter.iterselect = _wrap_on_obj(iterselect, adapter)
    patch_dialect(adapter.dialect)


def patch_dialect(dialect):
    _create_table_map = {
        'mysql': _create_table_mysql,
        'firebird': _create_table_firebird
    }
    dialect.create_table = _wrap_on_obj(
        _create_table_map.get(dialect.adapter.dbengine, _create_table), dialect
    )
    dialect.add_foreign_key_constraint = _wrap_on_obj(_add_fk_constraint, dialect)
    dialect.drop_constraint = _wrap_on_obj(_drop_constraint, dialect)


def patch_parser(dialect, parser):
    parser.registered['password'] = ParserMethodWrapper(
        parser,
        _parser_for_type('password')(_parser_password).f
    )
    parser.registered['reference'] = ParserMethodWrapper(
        parser,
        _parser_for_type('reference')(_parser_reference).f,
        parser._before_registry_['reference']
    )
    if 'geography' in dialect.types:
        parser.registered['geography'] = ParserMethodWrapper(
            parser,
            _parser_for_type('geography')(_parser_geo).f
        )
    if 'geometry' in dialect.types:
        parser.registered['geometry'] = ParserMethodWrapper(
            parser,
            _parser_for_type('geometry')(_parser_geo).f
        )


def patch_representer(representer):
    representer.registered_t['reference'] = TReprMethodWrapper(
        representer,
        _representer_for_type('reference')(_representer_reference),
        representer._tbefore_registry_['reference']
    )


def insert(adapter, table, fields):
    query = adapter._insert(table, fields)
    try:
        adapter.execute(query)
    except:
        e = sys.exc_info()[1]
        if hasattr(table, '_on_insert_error'):
            return table._on_insert_error(table, fields, e)
        raise e
    if not table._id:
        id = {
            field.name: val for field, val in fields
            if field.name in table._primarykey
        } or None
    elif table._id.type == 'id':
        id = adapter.lastrowid(table)
    else:
        id = {field.name: val for field, val in fields}.get(table._id.name)
    rid = typed_row_reference(id, table)
    return rid


def iterselect(adapter, query, fields, attributes):
    colnames, sql = adapter._select_wcols(query, fields, **attributes)
    return adapter.iterparse(sql, fields, colnames, **attributes)


def _expand_all_with_concrete_tables(adapter, fields, tabledict):
    new_fields, concrete_tables = [], []
    for item in fields:
        if isinstance(item, SQLALL):
            new_fields += item._table
            concrete_tables.append(item._table)
        elif isinstance(item, str):
            m = REGEX_TABLE_DOT_FIELD.match(item)
            if m:
                tablename, fieldname = m.groups()
                new_fields.append(adapter.db[tablename][fieldname])
            else:
                new_fields.append(Expression(adapter.db, lambda item=item: item))
        else:
            new_fields.append(item)
    # ## if no fields specified take them all from the requested tables
    if not new_fields:
        for table in tabledict.values():
            for field in table:
                new_fields.append(field)
            concrete_tables.append(table)
    return new_fields, concrete_tables


def _select_wcols(
    adapter,
    query,
    fields,
    left=False,
    join=False,
    distinct=False,
    orderby=False,
    groupby=False,
    having=False,
    limitby=False,
    orderby_on_limitby=True,
    for_update=False,
    outer_scoped=[],
    **kwargs
):
    return adapter._select_wcols_inner(
        query,
        fields,
        left=left,
        join=join,
        distinct=distinct,
        orderby=orderby,
        groupby=groupby,
        having=having,
        limitby=limitby,
        orderby_on_limitby=orderby_on_limitby,
        for_update=for_update,
        outer_scoped=outer_scoped
    )


def _select_aux(adapter, sql, fields, attributes, colnames):
    rows = adapter._select_aux_execute(sql)
    if isinstance(rows, tuple):
        rows = list(rows)
    limitby = attributes.get('limitby', None) or (0,)
    rows = adapter.rowslice(rows, limitby[0], None)
    return adapter.parse(
        rows,
        fields,
        colnames,
        concrete_tables=attributes.get('_concrete_tables', [])
    )


def parse(adapter, rows, fields, colnames, **options):
    fdata, tables = _parse_expand_colnames(adapter, fields)
    new_rows = [
        _parse(
            adapter,
            row,
            fdata,
            tables,
            options['concrete_tables'],
            fields,
            colnames,
            options.get('blob_decode', True)
        ) for row in rows
    ]
    rowsobj = adapter.db.Rows(adapter.db, new_rows, colnames, rawrows=rows)
    return rowsobj


def iterparse(adapter, sql, fields, colnames, **options):
    return IterRows(
        adapter.db,
        sql,
        fields,
        options.get('_concrete_tables', []),
        colnames
    )


def _parse_expand_colnames(adapter, fieldlist):
    rv, tables = [], {}
    for field in fieldlist:
        if not isinstance(field, Field):
            rv.append(None)
            continue
        table = field.table
        tablename, fieldname = table._tablename, field.name
        ft = field.type
        fit = field._itype
        rv.append((tablename, fieldname, table, field, ft, fit))
        tables[tablename] = table
    return rv, tables


def _parse(adapter, row, fdata, tables, concrete_tables, fields, colnames, blob_decode):
    new_row, rows_cls, rows_accum = _build_newrow_wtables(
        adapter, tables, concrete_tables
    )
    extras = adapter.db.Row()
    #: let's loop over columns
    for (idx, colname) in enumerate(colnames):
        value = row[idx]
        fd = fdata[idx]
        tablename = None
        #: do we have a real column?
        if fd:
            (tablename, fieldname, table, field, ft, fit) = fd
            colset = rows_accum[tablename]
            #: parse value
            value = adapter.parse_value(value, fit, ft, blob_decode)
            if field.filter_out:
                value = field.filter_out(value)
            colset[fieldname] = value
        #: otherwise we set the value in extras
        else:
            value = adapter.parse_value(
                value, fields[idx]._itype, fields[idx].type, blob_decode
            )
            extras[colname] = value
            new_column_name = adapter._regex_select_as_parser(colname)
            if new_column_name is not None:
                column_name = new_column_name.groups(0)
                new_row[column_name[0]] = value
    for key, val in rows_cls.items():
        new_row[key] = val._from_engine(rows_accum[key])
    #: add extras if needed (eg. operations results)
    if extras:
        new_row['_extra'] = extras
    return new_row


def _build_newrow_wtables(adapter, tables, concrete_tables):
    row, cls_map, accum = adapter.db.Row(), {}, {}
    for name, table in tables.items():
        cls_map[name] = adapter.db.Row
        accum[name] = {}
    for table in concrete_tables:
        cls_map[table._tablename] = table._model_._rowclass_
        accum[table._tablename] = {}
    return row, cls_map, accum


def _create_table(dialect, tablename, fields):
    return [
        "CREATE TABLE %s(\n    %s\n);" % (dialect.quote(tablename), fields)]


def _create_table_mysql(dialect, tablename, fields):
    return ["CREATE TABLE %s(\n    %s\n) ENGINE=%s CHARACTER SET utf8;" % (
        dialect.quote(tablename), fields,
        dialect.adapter.adapter_args.get('engine', 'InnoDB'))]


def _create_table_firebird(dialect, tablename, fields):
    rv = _create_table(dialect, tablename, fields)
    sequence_name = dialect.sequence_name(tablename)
    trigger_name = dialect.trigger_name(tablename)
    trigger_sql = (
        'create trigger %s for %s active before insert position 0 as\n'
        'begin\n'
        'if(new."id" is null) then\n'
        'begin\n'
        'new."id" = gen_id(%s, 1);\n'
        'end\n'
        'end;')
    rv.extend([
        'create generator %s;' % sequence_name,
        'set generator %s to 0;' % sequence_name,
        trigger_sql % (trigger_name, dialect.quote(tablename), sequence_name)
    ])
    return rv


def _add_fk_constraint(
    dialect,
    name,
    table_local,
    table_foreign,
    columns_local,
    columns_foreign,
    on_delete
):
    return (
        f"ALTER TABLE {dialect.quote(table_local)} "
        f"ADD CONSTRAINT {dialect.quote(name)} "
        f"FOREIGN KEY ({','.join([dialect.quote(v) for v in columns_local])}) "
        f"REFERENCES {dialect.quote(table_foreign)}"
        f"({','.join([dialect.quote(v) for v in columns_foreign])}) "
        f"ON DELETE {on_delete};"
    )


def _drop_constraint(dialect, name, table):
    return f"ALTER TABLE {dialect.quote(table)} DROP CONSTRAINT {dialect.quote(name)};"


def _parser_reference(parser, value, referee):
    if '.' not in referee:
        value = typed_row_reference(value, parser.adapter.db[referee])
    return value


def _parser_geo(parser, value):
    return GeoFieldWrapper(value)


def _parser_password(parser, value):
    return PasswordFieldWrapper(value)


def _representer_reference(representer, value, referenced):
    rtname, _, rfname = referenced.partition('.')
    rtable = representer.adapter.db[rtname]
    if not rfname and rtable._id:
        rfname = rtable._id.name
    if not rfname:
        return value
    rtype = rtable[rfname].type
    if isinstance(value, Row) and getattr(value, "_concrete", False):
        value = value[(value._model.primary_keys or ["id"])[0]]
    if rtype in ('id', 'integer'):
        return str(int(value))
    if rtype == 'string':
        return str(value)
    return representer.adapter.represent(value, rtype)


def _initialize(adapter, *args, **kwargs):
    adapter._find_work_folder()
    adapter._connection_manager.configure(
        max_connections=adapter.db._pool_size,
        connect_timeout=adapter.db._connect_timeout,
        stale_timeout=adapter.db._keep_alive_timeout)


def _begin(adapter):
    pass


def _in_transaction(adapter):
    return bool(adapter._connection_manager.state.transactions)


def _push_transaction(adapter, transaction):
    adapter._connection_manager.state.transactions.append(transaction)


def _pop_transaction(adapter):
    adapter._connection_manager.state.transactions.pop()


def _transaction_depth(adapter):
    return len(adapter._connection_manager.state.transactions)


def _top_transaction(adapter):
    if adapter._connection_manager.state.transactions:
        return adapter._connection_manager.state.transactions[-1]
